from unittest import TestCase

from pythautomata.automata_definitions.weighted_tomitas_grammars import WeightedTomitasGrammars
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.symbol import SymbolStr


class TestPDFALastTokenQueries(TestCase):

    def test_1(self):
        pdfa = WeightedTomitasGrammars.get_automaton_1()
        res = pdfa.get_last_token_weights(Sequence([]), [Sequence(
            [SymbolStr('$')]), Sequence([SymbolStr('0')]), Sequence([SymbolStr('1')])])
        self.assertTrue(True)

    def test_2(self):
        pdfa = WeightedTomitasGrammars.get_automaton_1()
        resTerminal = pdfa.last_token_weight(Sequence([SymbolStr('$')]))
        res0 = pdfa.last_token_weight(Sequence([SymbolStr('0')]))
        res1 = pdfa.last_token_weight(Sequence([SymbolStr('1')]))
        self.assertTrue(True)



import pydot
from pythautomata.automata.mealy_machine import MealyMachine
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.mealy_state import MealyState

from pythautomata.base_types.symbol import SymbolStr
from pythautomata.exceptions.model_importing_error import ModelImportingError
from pythautomata.model_comparators.mealy_machine_comparison_strategy import MealyMachineComparisonStrategy


class StandardDotNotationMealyImporter:

    @staticmethod
    def import_automata(path):
        graph = pydot.graph_from_dot_file(path)[0]
        name = graph.obj_dict['name']
        nodes = {}
        symbols = set()
        output_symbols = set()
        initial_state = None

        for node in graph.get_node_list():
            if node.get_name() not in ['__start0', '', '"\\n"']:
                nodes[node.get_name()] = MealyState(node.get_name())

        for edge in graph.get_edge_list():
            if edge.get_source() == '__start0':
                initial_state = nodes[edge.get_destination()]
            else:
                edge_label = edge.get_attributes()['label']
                if edge_label[0] == "\"" and edge_label[0] == edge_label[-1]:
                    edge_label = edge_label[1:-1]

                edge_label = edge_label.split("/")
                symbol = SymbolStr(edge_label[0])
                symbols.add(symbol)
                output_symbol = SymbolStr(edge_label[1])
                output_symbols.add(output_symbol)

                nodes[edge.get_source()].add_transition(
                    symbol, nodes[edge.get_destination()], output_symbol)

        states = nodes.values()
        alphabet = Alphabet(frozenset(symbols))
        output_alphabet = Alphabet(frozenset(output_symbols))
        if initial_state is None:
            raise ModelImportingError

        return MealyMachine(alphabet, output_alphabet, initial_state, states, MealyMachineComparisonStrategy(), name)



from math import exp
import math
from pythautomata.automata.moore_machine_automaton import MooreMachineAutomaton
from pythautomata.base_types.alphabet import Alphabet
from scipy.special import lambertw
import pythautomata.utilities.simple_mm_generator as simple_moore_machine_generator
from pythautomata.model_exporters.dot_exporters.moore_dot_exporting_strategy import MooreDotExportingStrategy


def generate_moore_machine(input_alphabet: Alphabet, output_alphabet: Alphabet, nominal_size: int, seed: int = None, exporting_strategies: list = [MooreDotExportingStrategy()]) -> MooreMachineAutomaton:
    """
    Function returning a randomly generated Moore Machine based on: 
        Nicaud, Cyril. (2014). Random Deterministic Automata. 5-23. 10.1007/978-3-662-44522-8_2. 

    Args:
        input_alphabet (Alphabet): Moore Machine input alphabet.
        output_alphabet (Alphabet): Moore Machine output alphabet.
        nominal_size (int): Target nominal size of the generated DFA.
        seed (int, optional): Random seed. Defaults to None.
        exporting_strategies (list, optional): List of exporting strategies. Defaults to [DotExportingStrategy()].

    Returns:
        MooreMachineAutomaton: Random Moore Machine with distribution of sizes centered near nominal_size and depth centered near 2*log(nominal_size,2)-2
    """
    k = len(input_alphabet)
    ro_k = k + lambertw(-k * exp(-k))
    ro_k = ro_k.real
    v_k = ro_k/k
    number_of_states = math.ceil(nominal_size / v_k)
    moore_machine = simple_moore_machine_generator.generate_moore_machine(
        input_alphabet, output_alphabet, number_of_states, seed=seed, exporting_strategies=exporting_strategies)
    return moore_machine



from abc import ABC, abstractmethod
import sys
from collections import namedtuple
from typing import Any, Optional, Union

from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.state import State
from pythautomata.base_types.moore_state import MooreState
from pythautomata.base_types.mealy_state import MealyState

ExecutionState = namedtuple("ExecutionState", "state sequence")


class FiniteAutomaton(ABC):
    _alphabet: Alphabet
    _exporting_strategies: list

    def __init__(self, comparator: 'FiniteAutomataComparator', calculate_access_strings: bool = False):
        self._comparator = comparator
        if calculate_access_strings:
            self.calculate_access_strings()

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def comparator(self) -> 'FiniteAutomataComparator':
        return self._comparator

    @comparator.setter
    def comparator(self, value: 'FiniteAutomataComparator'):
        self._comparator = value

    @property
    def alphabet(self) -> Alphabet:
        return self._alphabet

    @property
    def output_alphabet(self):
        return [True, False]

    @property
    @abstractmethod
    def hole(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def initial_states(self) -> frozenset:
        raise NotImplementedError

    @property
    def has_full_alphabet(self) -> bool:
        """Whether or not the automaton has the complete alphabet or not. For example, symbolic automaton might not have the full alphabet if it works with intervals

        Returns:
            bool: True by defauult, should be overriden if you want a different result
        """
        return True

    def find_first_difference_with(self, other: Any) -> Optional[Sequence]:
        return self.comparator.get_counterexample_between(self, other)
    
    def calculate_access_strings(self):
        for state in self.states:
            state.access_string = self.get_access_string(state)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, FiniteAutomaton) and self._comparator.are_equivalent(self, other)
    
    def get_access_string(self, target: Union[State, MooreState, MealyState]) -> Sequence:
        initial_state = list(self.initial_states)[0]
        if target == initial_state:
            return Sequence([])
        
        explored_states = set()
        queue = [(initial_state, [])]

        while queue:
            actual_state, seq = queue.pop(0)
            if actual_state not in explored_states:
                neighbours = set()
                for symbol, next_states in actual_state.transitions.items():
                    seq_copy = seq.copy()
                    seq_copy.append(SymbolStr(symbol))
                    next_state = list(next_states)[0]
                    if next_state == target:
                        return Sequence(seq_copy)

                    if next_state not in neighbours:
                        neighbours.add(next_state)
                        queue.append((next_state, seq_copy))
                    
            explored_states.add(actual_state)

        return Sequence([])

    def export(self, path=None) -> None:
        for strategy in self._exporting_strategies:
            try:
                strategy.export(self, path)
            except:
                print("Unexpected exception when exporting " +
                      str(self._name) + ": " + str(sys.exc_info()[0]))


class FiniteAutomataComparator(ABC):

    @abstractmethod
    def are_equivalent(self, model1: 'FiniteAutomaton', model2: 'FiniteAutomaton') -> bool:
        # Returns true iif both models are equivalent (i.e. they recongize the same
        # set of symbol sequences).
        raise NotImplementedError

    @abstractmethod
    def get_counterexample_between(self, model1: 'FiniteAutomaton', model2: 'FiniteAutomaton') -> Optional[Sequence]:
        # Should return a counterexample sequence (one that is recognized by one and only
        # one of both models) if parameters are not equivalent, or None if they are
        raise NotImplementedError

    @abstractmethod
    def equivalent_output(self, observation1, observation2) -> bool:
        # Returns true iif both observations are equivalent (according to the comparator criteria).
        raise NotImplementedError



from unittest import TestCase

from pythautomata.automata_definitions.weighted_tomitas_grammars import WeightedTomitasGrammars
from pythautomata.model_exporters.image_exporters.wfa_image_exporter import WFAImageExporter
from pythautomata.model_exporters.image_exporters.wfa_image_exporter_with_partition_mapper import WFAImageExporterWithPartitionMapper
from pythautomata.model_exporters.image_exporters.image_exporting_strategy import ImageExportingStrategy
from pythautomata.model_exporters.dot_exporters.wfa_dot_exporting_strategy import WFADotExportingStrategy
from pythautomata.utilities.probability_partitioner import TopKProbabilityPartitioner
from pythautomata.model_exporters.partition_mapper import TopKPartitionMapper


class TestWFAExporting(TestCase):
    def test_1(self):
        wfa = WeightedTomitasGrammars.get_automaton_7()
        WFADotExportingStrategy().export(wfa, "./output_models/tests")
        ImageExportingStrategy(WFADotExportingStrategy(), "png").export(
            wfa, "./output_models/tests")

        WFAImageExporter().export(wfa, "./output_models/tests")
        WFAImageExporterWithPartitionMapper(TopKProbabilityPartitioner(
            1), TopKPartitionMapper()).export(wfa, "./output_models/tests", wfa.name+"partitioned_top_1")



from abc import ABC, abstractmethod
from pythautomata.base_types.guard import Guard
from pythautomata.base_types.symbol import Symbol
from pythautomata.base_types.symbolic_state import SymbolicState

class BooleanAlgebraLearner(ABC):
    """Abstract class that boolean algebra learners should extend
    """
    @abstractmethod
    def learn(self, multidict:dict[SymbolicState, list[Symbol]]) -> list[tuple[Guard, SymbolicState]]:
        """[summary]

        Args:
            multidict (dict[SymbolicState, list[Symbol]]): a dictionary that maps a state to all the symbols with which the transition goes to said state

        Returns:
            list[tuple[Guard, SymbolicState]]: a list containing a guards and the state the guards transition to
        """
        pass


from abc import ABC, abstractmethod
from os import makedirs
from os.path import isdir
from pathlib import Path


class PDFAModelExportingStrategy(ABC):

    @abstractmethod
    def export(self, model, path: str):
        raise NotImplementedError

    def get_path_for(self, path: str, model, file_name: str):
        if path is None:
            name = model.name
            path = "output_models/" + \
                ("" if name is None else f"{name}")
        if not isdir(path):
            makedirs(path)
        if file_name is None:
            file_name = model.name
        return Path(path, file_name)



from graphviz import Digraph
from pythautomata.abstract.pdfa_model_exporting_strategy import PDFAModelExportingStrategy
from pythautomata.utilities.probability_partitioner import ProbabilityPartitioner
from pythautomata.model_exporters.partition_mapper import PartitionMapper
from collections import OrderedDict


class WFAImageExporterWithPartitionMapper(PDFAModelExportingStrategy):
    def __init__(self, partitioner: ProbabilityPartitioner, partition_mapper: PartitionMapper) -> None:
        super().__init__()
        self._partitioner = partitioner
        self._mapper = partition_mapper

    def export(self, model, path=None, file_name=None):
        graph = Digraph('weighted_automaton', format='png')
        graph.attr(rankdir='LR', margin='0', size='15')

        graph.attr('node', shape='circle')
        states = sorted(model.weighted_states, key=lambda x: str(x.name))
        for state in states:
            state_name = str(state.name)
            symbols_positions, weights, _ = state.get_all_symbol_weights()
            symb_weight = OrderedDict(zip(symbols_positions, weights))
            state_partition = self._partitioner.get_partition(weights)
            label_partition = self._mapper.get_str_for_partition(symb_weight,
                                                                 state_partition)
            label = state_name + "\n" + label_partition
            if state.initial_weight == 1:
                graph.node(state_name, label, shape='diamond')
            else:
                graph.node(state_name, label)
            transitions = dict()
            for symbol, weighted_transitions in state.transitions_list.items():
                for weighted_transition in weighted_transitions:
                    weighted_transition_name = str(weighted_transition[0].name)
                    transition_representation = self._mapper.get_str_for_transition(
                        symb_weight, symbol, state_partition)
                    if (state_name, weighted_transition_name) in transitions.keys():
                        new = transitions[(state_name, weighted_transition_name)] + "\n" + str(symbol) + "-" + \
                            transition_representation
                    else:
                        new = str(symbol) + "-" + transition_representation
                    transitions[(state_name, weighted_transition_name)] = new
            for key in transitions.keys():
                s_from, s_to = key
                graph.edge(s_from, s_to, str(transitions[key]))

        graph.render(self.get_path_for(path, model, file_name),
                     cleanup=True, format='pdf')



from abc import ABC, abstractmethod
from typing import Union
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence

class Model(ABC):

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError


    @property
    @abstractmethod
    def alphabet(self) -> Alphabet:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def output_alphabet(self) -> Alphabet:
        raise NotImplementedError
    
    @abstractmethod
    def process_query(self, seq: Sequence) -> Union[Sequence, bool]:
        raise NotImplementedError


import uuid
from pythautomata.abstract.boolean_model import BooleanModel
from pythautomata.base_types.state import State
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence
from collections import deque, namedtuple
from pythautomata.abstract.finite_automaton import FiniteAutomaton, FiniteAutomataComparator
from pythautomata.model_exporters.encoded_file_exporting_strategy import EncodedFileExportingStrategy

ExecutionState = namedtuple("ExecutionState", ["state", "sequence"])


class NondeterministicFiniteAutomaton(FiniteAutomaton, BooleanModel):
    """
    Implementation of NFA.

    Attributes
    ----------
    states: set[State]
        Set containing the DFA's states
    initial_states: frozenset[State]
        Set containing all initial states of the NFA. Also included in "states"
    """

    def __init__(self, alphabet: Alphabet, initial_states: frozenset[State], states: set[State],
                 comparator: FiniteAutomataComparator, name: str = None,
                 exportingStrategies: list = [EncodedFileExportingStrategy()], hole: State = State("Hole")):
        self.states = states
        for state in self.states:
            assert all(symbol in alphabet for symbol in state.transitions)
            state.add_hole_transition(hole)

        self._name = 'NFA - ' + str(uuid.uuid4().hex) if name is None else name
        self._alphabet = alphabet
        self._initial_states = initial_states
        self._set_hole(hole)
        self._exporting_strategies = exportingStrategies
        super(NondeterministicFiniteAutomaton, self).__init__(comparator)

    def accepts(self, sequence: Sequence) -> bool:
        toVisit = deque(ExecutionState(state, sequence)
                        for state in self._initial_states)
        while toVisit:
            executionState = toVisit.pop()
            state = executionState.state
            value = executionState.sequence.value
            if len(value) > 0:
                suffix = Sequence(value[1:])
                for destination in state.next_states_for(value[0]):
                    if destination != self.hole:
                        toVisit.appendleft(ExecutionState(destination, suffix))
            elif state.is_final:
                return True
        return False

    def process_query(self, sequence: Sequence) -> bool:
        return self.accepts(sequence)

    @property
    def initial_states(self) -> frozenset:
        return self._initial_states

    @property
    def hole(self):
        return self._hole

    def _set_hole(self, hole: State) -> None:
        self._hole = hole
        # hole's hole is itself
        self._hole.add_hole_transition(self.hole)



from abc import ABC, abstractmethod

from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence


class BooleanModel(ABC):

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def alphabet(self) -> Alphabet:
        raise NotImplementedError

    @abstractmethod
    def accepts(self, sequence: Sequence) -> bool:
        raise NotImplementedError



import random
from random import randint

from pythautomata.abstract.finite_automaton import FiniteAutomaton
from pythautomata.base_types.sequence import Sequence

class StatePrefixRandomWalkComparisonStrategy:

    def  __init__(self, number_steps, reset_probability):
        self.steps = number_steps
        self.reset_prob = reset_probability

    def are_equivalent(self, model1, model2):
        return self.get_counterexample_between(model1, model2) is None

    def get_counterexample_between(self, hypothesis: FiniteAutomaton, oracle):
        symbols = list(hypothesis.alphabet.symbols)
        symbols.sort()
        
        for state in hypothesis.states:
            prefix = state.access_string
            if prefix is None:
                prefix = hypothesis.get_access_string(state)

            suffix = Sequence()
            for _ in range(self.steps):
                pos = randint(0, len(symbols) - 1)
                suffix = suffix.append(symbols[pos])

                if hypothesis.process_query(prefix+suffix) != oracle.process_query(prefix+suffix):
                    return prefix+suffix

                if random.random() <= self.reset_prob:
                    suffix = Sequence()
                
        return None



import numpy as np
import random
import math

from pythautomata.base_types.symbol import Symbol
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.alphabet import Alphabet
from typing import List
from pythautomata.utilities.sequence_generator import SequenceGenerator


class UniformWordSequenceGenerator(SequenceGenerator):

    def __init__(self, alphabet: Alphabet, max_seq_length: int, random_seed: int = 21, min_seq_length: int = 0):
        super().__init__(alphabet, max_seq_length, random_seed, min_seq_length)
        self._probs = {x:self._prob_by_word_length(x) for x in range(min_seq_length, max_seq_length)}
    
    def generate_words(self, number_of_words: int):
        result =[]
        for _ in range(number_of_words):
            length = self._select_random_length()
            result.append(self.generate_single_word(length))
        return result

    def generate_single_word(self, length):
        if length > self._max_seq_length:
            raise AssertionError("Param length cannot exceed max_seq_length")
        if length < self._min_seq_length:
            raise AssertionError("Param length cannot be less than max_seq_length")

        value = []
        list_symbols = list(self._alphabet.symbols)
        list_symbols.sort()
        for _ in range(length):
            position = random.randint(0, len(list_symbols) - 1)
            symbol = list_symbols[position]
            value.append(symbol)
        return Sequence(value)

    def _prob_by_word_length(self, length):
        s = len(self._alphabet)
        n = self._max_seq_length
        m = self._min_seq_length

        si = math.pow(s,length)
        sn_plus_1 = math.pow(s,n+1)
        sm = math.pow(s, m)

        proba = si*(s-1)/(sn_plus_1-sm)
        return proba

    def _select_random_length(self):
        length = random.choices(list(self._probs.keys()), weights = self._probs.values(),k=1)[0]
        return length

    


from pythautomata.abstract.probabilistic_model import ProbabilisticModel
from pythautomata.base_types.sequence import Sequence
from pythautomata.utilities import pdfa_utils
import math
from sklearn.metrics import ndcg_score
import numpy as np
epsilon = Sequence()


def ndcg_score_avg(target_model: ProbabilisticModel, learned_model: ProbabilisticModel, test_sequences: list[Sequence]):
    suffixes = list()
    suffixes.append(Sequence() + learned_model.terminal_symbol)
    ndcg = list()
    k = math.ceil(len(learned_model.alphabet) / 2)

    for symbol in learned_model.alphabet.symbols:
        suffixes.append(Sequence((symbol,)))

    all_obs1 = target_model.last_token_probabilities_batch(
        test_sequences, suffixes)
    all_obs2 = learned_model.last_token_probabilities_batch(
        test_sequences, suffixes)

    for i in range(len(all_obs1)):
        obs1 = np.asarray([all_obs1[i]])
        obs2 = np.asarray([all_obs2[i]])
        ndcg_word_score = ndcg_score(obs1, obs2, k=k)
        ndcg.append(ndcg_word_score)

    return np.mean(np.array(ndcg))


def wer_avg(target_model: ProbabilisticModel, learned_model: ProbabilisticModel, test_sequences: list[Sequence]):
    suffixes = list()
    suffixes.append(Sequence() + learned_model.terminal_symbol)
    wer = 0

    for symbol in learned_model.alphabet.symbols:
        suffixes.append(Sequence((symbol,)))

    all_obs1 = target_model.last_token_probabilities_batch(
        test_sequences, suffixes)
    all_obs2 = learned_model.last_token_probabilities_batch(
        test_sequences, suffixes)

    for i in range(len(all_obs1)):
        obs1 = np.asarray(all_obs1[i])
        obs2 = np.asarray(all_obs2[i])
        max1 = np.argmax(obs1)
        max2 = np.argmax(obs2)
        if max1 != max2:
            wer += 1

    return wer / len(test_sequences)


def log_probability_error(target_model: ProbabilisticModel, learned_model: ProbabilisticModel, test_sequences: list[Sequence]):
    errors = list()
    for seq in test_sequences:
        errors.append(abs(target_model.log_sequence_weight(seq) -
                      learned_model.log_sequence_weight(seq)))
    return np.mean(np.array(errors))


def out_of_tolerance_elements(target_model: ProbabilisticModel, learned_model: ProbabilisticModel, test_sequences: list[Sequence], tolerance):
    suffixes = list()
    errorCount = 0
    suffixes.append(Sequence() + learned_model.terminal_symbol)
    for symbol in learned_model.alphabet.symbols:
        suffixes.append(Sequence((symbol,)))

    all_obs1 = target_model.last_token_probabilities_batch(
        test_sequences, suffixes)
    all_obs2 = learned_model.last_token_probabilities_batch(
        test_sequences, suffixes)

    for i in range(len(all_obs1)):
        obs1 = np.asarray(all_obs1[i])
        obs2 = np.asarray(all_obs2[i])
        if not pdfa_utils.are_within_tolerance_limit(obs1, obs2, tolerance):
            errorCount += 1

    return errorCount


def out_of_partition_elements(target_model: ProbabilisticModel, learned_model: ProbabilisticModel, test_sequences: list[Sequence], partitions: int):
    suffixes = list()
    errorCount = 0
    suffixes.append(Sequence() + learned_model.terminal_symbol)
    for symbol in learned_model.alphabet.symbols:
        suffixes.append(Sequence((symbol,)))

    all_obs1 = target_model.last_token_probabilities_batch(
        test_sequences, suffixes)
    all_obs2 = learned_model.last_token_probabilities_batch(
        test_sequences, suffixes)

    for i in range(len(all_obs1)):
        obs1 = np.asarray(all_obs1[i])
        obs2 = np.asarray(all_obs2[i])
        p1 = pdfa_utils.get_quantized_interval_partitions(obs1, partitions)
        p2 = pdfa_utils.get_quantized_interval_partitions(obs2, partitions)
        result = pdfa_utils.compare_quantized_interval_partition(p1, p2)
        if not result:
            errorCount += 1

    return errorCount


def absolute_error_avg(target_model: ProbabilisticModel, learned_model: ProbabilisticModel, test_sequences: list[Sequence]):
    suffixes = list()
    suffixes.append(Sequence() + learned_model.terminal_symbol)

    for symbol in learned_model.alphabet.symbols:
        suffixes.append(Sequence((symbol,)))

    all_obs1 = target_model.last_token_probabilities_batch(
        test_sequences, suffixes)
    all_obs2 = learned_model.last_token_probabilities_batch(
        test_sequences, suffixes)

    absolute_error_sum = 0
    for i in range(len(all_obs1)):
        obs1 = np.asarray(all_obs1[i])
        obs2 = np.asarray(all_obs2[i])
        absolute_error_sum += np.sum(np.abs(obs1-obs2))

    return absolute_error_sum / len(test_sequences)


def cross_entropy(target_distribution, learned_distribution):
    # Categorical Cross Entropy Metric implemented following Keras implementation
    y_true = target_distribution
    y_pred = learned_distribution

    mask_0 = y_pred == 0
    y_true = y_true + mask_0*1e-7
    y_pred = y_pred + mask_0*1e-7

    mask_1 = y_pred == 1
    y_true = y_true - mask_1*1e-7
    y_pred = y_pred - mask_1*1e-7

    y_pred = y_pred/np.sum(y_pred)
    cross_entropy = - np.sum(y_true*np.log(y_pred))
    return cross_entropy


def mean_cross_entropy(target_model: ProbabilisticModel, learned_model: ProbabilisticModel, test_sequences: list[Sequence]):
    suffixes = list()
    suffixes.append(Sequence() + learned_model.terminal_symbol)
    for symbol in learned_model.alphabet.symbols:
        suffixes.append(Sequence((symbol,)))

    target_obs = target_model.last_token_probabilities_batch(
        test_sequences, suffixes)
    learned_obs = learned_model.last_token_probabilities_batch(
        test_sequences, suffixes)

    ce_sum = 0
    for i in range(len(target_obs)):
        obs1 = np.asarray(target_obs[i])
        obs2 = np.asarray(learned_obs[i])
        ce_sum += cross_entropy(obs1, obs2)

    return ce_sum / len(test_sequences)



from unittest import TestCase

from pythautomata.automata_definitions.tomitas_grammars import TomitasGrammars
from pythautomata.model_exporters.standard_exporters.dfa_standard_dot_exporting_strategy import DfaStandardDotExportingStrategy
from pythautomata.utilities.standard_dot_notation_importer import StandardDotNotationImporter


class TestDFALoading(TestCase):

    def test_1(self):
        for dfa in TomitasGrammars.get_all_automata():
            name = dfa.name
            dfa.name = name + "_standard"
            DfaStandardDotExportingStrategy().export(dfa, "./output_models/tests")
            path = './output_models/tests'+'/'+dfa.name+'.dot'
            loaded_dfa = StandardDotNotationImporter.import_automata(path)
            self.assertEqual(dfa, loaded_dfa)



from random import seed
from random import randint

from pythautomata.base_types.symbol import Symbol
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.alphabet import Alphabet
from typing import List
from pythautomata.utilities.sequence_generator import SequenceGenerator


class UniformLengthSequenceGenerator(SequenceGenerator):

    def __init__(self, alphabet: Alphabet, max_seq_length: int, random_seed: int = 21, min_seq_length: int = 0):
        super().__init__(alphabet, max_seq_length, random_seed, min_seq_length)
    
    def generate_words(self, number_of_words: int):
        result = []
        for _ in range(number_of_words):
            length = randint(self._min_seq_length, self._max_seq_length)
            result.append(self.generate_single_word(length))
        return result

    def generate_single_word(self, length):
        if length > self._max_seq_length:
            raise AssertionError("Param length cannot exceed max_seq_length")

        value = []
        list_symbols = list(self._alphabet.symbols)
        list_symbols.sort()
        for _ in range(length):
            position = randint(0, len(list_symbols) - 1)
            symbol = list_symbols[position]
            value.append(symbol)
        return Sequence(value)

    


from pythautomata.base_types.sequence import Sequence
from pythautomata.automata.wheighted_automaton_definition.probabilistic_deterministic_finite_automaton import \
    ProbabilisticDeterministicFiniteAutomaton as PDFA
import numpy as np


def get_representative_sample(pdfa: PDFA, sample_size: int):
    assert (sample_size >= 0)
    sample = list()
    for i in range(sample_size):
        sample.append(_get_representative_word(pdfa))
    return sample


def _get_representative_word(pdfa: PDFA):
    word = Sequence()
    first_state = list(filter(lambda x: x.initial_weight ==
                       1, pdfa.weighted_states))[0]
    symbols, weights, next_states = first_state.get_all_symbol_weights()
    next_symbol = np.random.choice(symbols, p=weights)
    while next_symbol != pdfa.terminal_symbol:
        word += next_symbol
        i = symbols.index(next_symbol)
        next_state = next_states[i]
        symbols, weights, next_states = next_state.get_all_symbol_weights()
        next_symbol = np.random.choice(symbols, p=weights)
    return word



from pythautomata.abstract.model_exporting_strategy import \
    ModelExportingStrategy
from pythautomata.automata.deterministic_finite_automaton import \
    DeterministicFiniteAutomaton as DFA
from pythautomata.automata.mealy_machine import MealyMachine
from pythautomata.automata.moore_machine_automaton import \
    MooreMachineAutomaton as MooreMachine
from pythautomata.automata.non_deterministic_finite_automaton import \
    NondeterministicFiniteAutomaton as NFA
from pythautomata.automata.symbolic_finite_automaton import \
    SymbolicFiniteAutomaton as SFA
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.guard import Guard
from pythautomata.base_types.mealy_state import MealyState
from pythautomata.base_types.state import State
from pythautomata.base_types.symbol import Symbol, SymbolStr
from pythautomata.base_types.symbolic_state import SymbolicState
from pythautomata.base_types.moore_state import MooreState
from pythautomata.boolean_algebra_learner.boolean_algebra_learner import \
    BooleanAlgebraLearner as BAL
from pythautomata.model_comparators.dfa_comparison_strategy import \
    DFAComparisonStrategy as DFAComparator
from pythautomata.model_comparators.mealy_machine_comparison_strategy import MealyMachineComparisonStrategy
from pythautomata.model_comparators.moore_machine_comparison_strategy import MooreMachineComparisonStrategy


class AutomataConverter():\

    @staticmethod
    def convert_nfa_to_dfa(non_deterministic_finite_automaton: NFA) -> DFA:
        """
        Converts a given non deterministic finite automaton into a deterministic finite automaton.

        Args:
            non_deterministic_finite_automaton (NFA): Input non deterministic finite automaton.

        Returns:
            DFA: DFA equivalent to the inputted NFA.
        """
        initial_states = list(
            non_deterministic_finite_automaton.initial_states)
        symbols = list(non_deterministic_finite_automaton.alphabet.symbols)
        next_states_after_initial = AutomataConverter._get_next_states_from_state(
            non_deterministic_finite_automaton, initial_states)
        new_transitions: dict[frozenset[State], list[list[State]]] = {
            frozenset(initial_states): next_states_after_initial}

        # TODO extract func
        completed = False
        while not completed:
            for states in new_transitions.copy().keys():
                for next_states in new_transitions[states]:
                    if len(next_states) > 0 and not frozenset(next_states) in new_transitions:
                        next_states_after_this = AutomataConverter._get_next_states_from_state(
                            non_deterministic_finite_automaton, next_states)
                        new_transitions[frozenset(
                            next_states)] = next_states_after_this
            completed = True

            for next_states_by_symbol in new_transitions.values():
                for next_states in next_states_by_symbol:
                    if len(next_states) > 0 and not frozenset(next_states) in new_transitions.keys():
                        completed = False
                        break

        # TODO extract func
        new_states_pairs = []
        new_initial_state = None
        for states in new_transitions.keys():
            stateName = " and ".join([state.name for state in states])
            is_final = any([state.is_final for state in states])
            new_state = State(stateName, is_final)
            if states == set(initial_states):
                new_initial_state = new_state
            new_states_pairs.append((new_state, states))

        # TODO extract func
        for (new_state, dict_key) in new_states_pairs:
            for ind_symbol, next_states in enumerate(new_transitions[dict_key]):
                if len(next_states) > 0:
                    symbol = symbols[ind_symbol]

                    next_new_state = AutomataConverter._find_state(
                        next_states, new_states_pairs)

                    if next_new_state is not None:
                        new_state.add_transition(symbol, next_new_state)

        comparator = DFAComparator()
        return DFA(
            non_deterministic_finite_automaton.alphabet,
            new_initial_state,
            set((new_state for (new_state, dict_key) in new_states_pairs)),
            comparator=comparator
        )

    @staticmethod
    def _find_state(next_states, new_states_pairs):
        for (new_state, dict_key) in new_states_pairs:
            if dict_key == frozenset(next_states):
                return new_state

    @staticmethod
    def _get_next_states_from_state(automaton, states: list[State]) -> list[list[State]]:
        symbols = automaton.alphabet.symbols
        result: list[list[State]] = [[] for sym in symbols]

        for state in states:
            for sym_index, symbol in enumerate(symbols):
                next_states = state.next_states_for(symbol)
                if automaton.hole not in next_states:
                    result[sym_index].extend(next_state for next_state in next_states
                                             if next_state not in result[sym_index])

        return result

    @staticmethod
    def convert_dfa_to_sfa(dfa: DFA, b_a_learner: BAL, exportingStrategies: list[ModelExportingStrategy] = []) -> SFA:
        """
        Converts a given deterministic finite automaton into a symbolic finite automaton.

        Args:
            automaton (DFA): Input deterministic finite automaton.

        Returns:
            SFA: SFA equivalent to the inputted DFA.
        """
        initial_state: SymbolicState
        states: dict[str, SymbolicState] = {}
        # get all states and convert them to symbolic state
        initial_state = None
        for state in dfa.states:
            new_state: SymbolicState = SymbolicState(
                state.name, state.is_final)
            if new_state.name == dfa.initial_state.name:
                initial_state = new_state
            states[new_state.name] = new_state

        for state in dfa.states:
            # for each state, get the multimap that has the next state as key, and a list of symbols as value
            multidict: dict[SymbolicState, list[Symbol]] = {}
            for symbol, state_set in state.transitions.items():
                current_state = list(state_set).pop()
                transition_state = states[current_state.name]
                if transition_state in multidict:
                    multidict[transition_state].append(symbol)
                else:
                    multidict[transition_state] = [symbol]
            guard_state_list: list[tuple[Guard, SymbolicState]
                                   ] = b_a_learner.learn(multidict)
            for guard, s in guard_state_list:
                states[state.name].add_transition(guard, s)

        name = None if dfa.name == None else "SFA_"+dfa.name
        return SFA(dfa.alphabet, initial_state, set(states.values()), name, exportingStrategies)

    @staticmethod
    def convert_dfa_to_moore_machine(dfa: DFA) -> MooreMachine:
        """
        Converts a given deterministic finite automaton into a moore machine.

        Args:
            automaton (DFA): Input deterministic finite automaton.

        Returns:
            MooreMachine: MooreMachine equivalent to the inputted DFA.
        """
        initial_state: MooreState
        states: dict[str, MooreState] = {}

        # create alphabet with values = {0,1}
        output_alphabet = Alphabet(
            frozenset((SymbolStr('False'), SymbolStr('True'))))

        # get all states and convert them to moore machine state
        initial_state = None
        for state in dfa.states:
            if state.is_final:
                new_state = MooreState(
                    state.name, output_alphabet['True']
                )
            else:
                new_state = MooreState(
                    state.name, output_alphabet['False']
                )
            if new_state.name == dfa.initial_state.name:
                initial_state = new_state
            states[new_state.name] = new_state

        for state in dfa.states:
            # for each state, get all the transitions
            for symbol, state_set in state.transitions.items():
                current_state = list(state_set).pop()
                transition_state = states[current_state.name]
                states[state.name].add_transition(symbol, transition_state)

        name = None if dfa.name == None else "MooreMachine_"+dfa.name

        hole_state = MooreState("Hole", SymbolStr("False"))

        return MooreMachine(dfa.alphabet, output_alphabet, initial_state, set(states.values()), MooreMachineComparisonStrategy(), name=name, hole=hole_state)

    @staticmethod
    def convert_moore_machine_to_mealy_machine(moore_machine: MooreMachine) -> MealyMachine:
        """
        Converts a given moore machine into a mealy machine.

        Args:
            moore_machine (MooreMachine): Input moore machine.

        Returns:
            MealyMachine: MealyMachine equivalent to the inputted MooreMachine.
        """
        initial_state: MealyState
        states: dict[str, MealyState] = {}

        # get all states and convert them to output function λ'(q, a) = λ(δ(q, a))
        # where λ is the output function of the Moore machine and λ' is the output function of the Mealy machine
        # and δ is the transition function of the Moore machine
        initial_state = None

        for state in moore_machine.states:
            new_state = MealyState(state.name)
            if new_state.name == moore_machine.initial_state.name:
                initial_state = new_state
            states[new_state.name] = new_state

        for state in moore_machine.states:
            for symbol, state_set in state.transitions.items():
                current_state = list(state_set).pop()
                states[state.name].add_transition(
                    symbol, states[current_state.name], current_state.value)

        name = None if moore_machine._name == None else "MooreMachine_"+moore_machine._name

        hole_state = MealyState("Hole")

        return MealyMachine(moore_machine._alphabet, moore_machine.output_alphabet,
                            initial_state,
                            set(states.values()), MealyMachineComparisonStrategy(),
                            name=name, hole=hole_state)
    
    @staticmethod
    def convert_moore_machine_to_minimal_mealy_machine(moore_machine: MooreMachine) -> MealyMachine:
        """
        Converts a given moore machine into a mealy machine.

        Args:
            moore_machine (MooreMachine): Input moore machine.

        Returns:
            MealyMachine: MealyMachine equivalent to the inputted MooreMachine.
        """
        initial_state: MealyState = None
        states: dict[tuple, MealyState] = {}
        mealy_table: dict[tuple, str] = {}
        state_map: dict[str, str] = {}

        ordered_alphabet = list(moore_machine.alphabet.symbols)

        for state in moore_machine.states:
            transition_list = []
            for symbol in ordered_alphabet:
                next_state = state.transitions[symbol].pop()
                transition_list.append((symbol, next_state.value, next_state.name))
            transition_tuple = tuple(transition_list)
            if transition_tuple not in mealy_table:
                mealy_table[transition_tuple] = state.name
                state_map[state.name] = state.name
            else:
                state_map[state.name] = mealy_table[transition_tuple]

        for name in mealy_table.values():
            states[name] = MealyState(name)

        for state_transitions, state_name in mealy_table.items():
            for symbol, output, moore_next_state in state_transitions:
                state = states[state_name]
                # We store the next moore state, which has to be mapped to the mealy state
                # Using the state_map. This cannot be done earlier because we need to compelete the mealy table
                # in order to know all unique Mealy states and which moore states they represent.
                next_state = states[state_map[moore_next_state]]
                state.add_transition(symbol, next_state, output)

        initial_state = states[state_map[moore_machine.initial_state.name]]

        name = None if moore_machine._name == None else "MooreMachine_"+moore_machine._name

        hole_state = MealyState("Hole")

        return MealyMachine(moore_machine._alphabet, moore_machine.output_alphabet,
                            initial_state,
                            set(states.values()), MealyMachineComparisonStrategy(),
                            name=name, hole=hole_state)



from abc import ABC, abstractmethod
from collections import OrderedDict
from pythautomata.base_types.symbol import Symbol
import numpy as np


class PartitionMapper(ABC):
    @abstractmethod
    def get_str_for_partition(self, symbols_and_probabilities: OrderedDict, partition):
        raise NotImplementedError

    @abstractmethod
    def get_str_for_transition(self, symbols_and_probabilities: OrderedDict, symbol: Symbol, partition):
        raise NotImplementedError


class TopKPartitionMapper(PartitionMapper):
    def get_str_for_partition(self, symbols_and_probabilities: OrderedDict, partition):
        symbols = list(symbols_and_probabilities.keys())
        top_k_symbols = np.array(symbols)[partition == 1]
        return str(top_k_symbols)

    def get_str_for_transition(self, symbols_and_probabilities: OrderedDict, symbol: Symbol, partition):
        symbols = list(symbols_and_probabilities.keys())
        top_k_symbols = np.array(symbols)[partition == 1]
        return str(symbol in top_k_symbols)



import sys
import uuid

from pythautomata.abstract.model import Model
from pythautomata.abstract.finite_automaton import FiniteAutomaton
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.mealy_state import MealyState
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.symbol import Symbol
from pythautomata.exceptions.non_deterministic_states_exception import NonDeterministicStatesException
from pythautomata.exceptions.unknown_symbols_exception import UnknownSymbolsException
from pythautomata.model_exporters.standard_exporters.mealy_standard_dot_exporting_strategy import MealyStandardDotExportingStrategy

class MealyMachine(Model, FiniteAutomaton):
    """
    Implementation of Mealy Machines.

    Attributes
    ----------
    states: set[MealyState]
        Set of states of the machine.
    initial_state: MealyState
        Initial state of the machine. Also included in "states".
    input_alphabet: Alphabet
        Set of Symbols that defines the Symbols that can be used as inputs.
    output_alphabet: Alphabet
        Set of Symbols that defines the Symbols that can be used as output.
    name: str
        Name of the machine.
    exporting_strategies: list
        List of exporting strategies that can be used to export the machine.
    """

    def __init__(self, input_alphabet: Alphabet, output_alphabet: Alphabet,
                 initial_state: MealyState, states: set[MealyState], comparator,
                 name: str = None, exporting_strategies: list = [MealyStandardDotExportingStrategy()],
                 hole: MealyState = MealyState("\u22A5"),
                 calculate_access_strings: bool = False):
        self.states = states
        for state in self.states:
            self._verify_state(state, input_alphabet, output_alphabet)
            state.add_hole_transition(hole)

        self._name = ("Mealy Machine - " + str(uuid.uuid4().hex)
                      if name is None else name)
        self._alphabet = input_alphabet
        self._output_alphabet = output_alphabet
        self.initial_state = initial_state
        self._set_hole(hole)
        self._exporting_strategies = exporting_strategies
        self._comparator = comparator
        self._actual_state = initial_state
        super(MealyMachine, self).__init__(comparator, calculate_access_strings)

    @property
    def initial_states(self) -> frozenset:
        return frozenset([self.initial_state])

    @property
    def hole(self):
        return self._hole
    
    @property
    def name(self) -> Alphabet:
        return self._name

    @property
    def alphabet(self) -> Alphabet:
        return self._alphabet

    @property
    def output_alphabet(self) -> Alphabet:
        return self._output_alphabet

    def __eq__(self, other):
        return isinstance(other, MealyMachine) and self._comparator.are_equivalent(self, other)

    def step(self, symbol):
        output = self._actual_state.outputs[symbol]
        self._actual_state = self._actual_state.next_state_for(symbol)

        return output

    def reset(self):
        self._actual_state = self.initial_state

    def last_symbol(self, sequence: Sequence) -> Symbol:
        actual_state = self.initial_state
        output = None
        for symbol in sequence.value:
            output = actual_state.outputs[symbol]
            actual_state = actual_state.next_state_for(symbol)
        return output

    def process_query(self, sequence: Sequence) -> Symbol:
        return self.last_symbol(sequence)

    def transduce(self, sequence: Sequence) -> Sequence:
        actual_state = self.initial_state
        output = Sequence()
        for symbol in sequence.value:
            output.append(actual_state.outputs[symbol])
            actual_state = actual_state.next_state_for(symbol)
        return output

    def export(self, path=None):
        for strategy in self._exporting_strategies:
            try:
                strategy.export(self, path)
            except:
                print("Unexpected exception when exporting " +
                      str(self._name) + ": " + str(sys.exc_info()[0]))

    def _set_hole(self, hole: MealyState) -> None:
        self._hole = hole
        self.hole.add_hole_transition(hole)

    def _verify_state(self, state: MealyState, input_alphabet: Alphabet,
                      output_alphabet: Alphabet) -> None:
        self.verify_transitions(state, input_alphabet)
        self.verify_output(state, output_alphabet)
        self.verify_state_is_deterministic(state)

    def verify_transitions(self, state: MealyState, input_alphabet: Alphabet):
        for symbol in state.transitions:
            if symbol not in input_alphabet:
                raise UnknownSymbolsException()

    def verify_output(self, state: MealyState, output_alphabet: Alphabet):
        for _, symbol in state.outputs.items():
            if symbol not in output_alphabet:
                raise UnknownSymbolsException()

    def verify_state_is_deterministic(self, state: MealyState):
        if not state.is_deterministic:
            raise NonDeterministicStatesException()



from abc import abstractmethod
from typing import Optional

from pythautomata.abstract.finite_automaton import FiniteAutomataComparator
from pythautomata.base_types.sequence import Sequence
from pythautomata.automata.wheighted_automaton_definition.weighted_state import WeightedState


class WFAComparator(FiniteAutomataComparator):
    """
    Class containing a WFA Comparator Strategy based on next symbol distributions and a quantization parameter.

    Methods
    -------
    are_equivalent: bool
        returns true iif wfa1 is equivalent to wfa2 according to which partitions the distributions over next symbols belong.

    get_counterexample_between: Sequence
        returns a Sequence where the next token weights differ (or belong to different partitions)
    """

    def __init__(self) -> None:
        super().__init__()

    def are_equivalent(self, wfa1, wfa2) -> bool:
        return self.get_counterexample_between(wfa1, wfa2) is None

    def get_counterexample_between(self, wfa1, wfa2) -> Optional[Sequence]:
        if wfa1.alphabet != wfa2.alphabet:
            raise ValueError("Alphabets are not equivalent.")
        alphabet = wfa1.alphabet
        # initialPairs is a pair of initial states
        initial_pair: tuple[WeightedState, WeightedState] = (
            min(wfa1.initial_states), min(wfa2.initial_states))

        pairs_to_visit = [initial_pair]
        sequence_for_pairs = {initial_pair: Sequence()}
        visited_pairs: set[tuple[WeightedState, WeightedState]] = set()

        while pairs_to_visit:
            pair = pairs_to_visit[0]
            if not self._states_are_equivalent(pair[0], pair[1], alphabet):
                return sequence_for_pairs[pair]
            for symbol in alphabet.symbols:
                self._process_equivalence_iteration_with(symbol, pairs_to_visit,
                                                         visited_pairs, sequence_for_pairs)
            # pairs_to_visit.remove(pair)
            pairs_to_visit = list(
                filter(lambda x: not self._pair_equivalent_by_name(x, pair), pairs_to_visit))
            visited_pairs.add(pair)
        return None

    @abstractmethod
    def equivalent_output(self, observation1, observation2) -> bool:
        "Asumes observation is any collection of weights"
        raise NotImplementedError

    @abstractmethod
    def next_tokens_equivalent_output(self, observation1, observation2) -> bool:
        "Asumes observation is any collection of weights corresponding to the next token weights of a state"
        raise NotImplementedError

    def _pair_equivalent_by_name(self, states_pair1, states_pair2):
        return states_pair1[0].name == states_pair2[0].name and states_pair1[1].name == states_pair2[1].name

    def _states_are_equivalent(self, state1, state2, alphabet, ignore_initial_weight=True):
        r1 = self.equivalent_values(
            state1.initial_weight, state2.initial_weight)
        r2 = self.equivalent_values(state1.final_weight, state2.final_weight)
        r3 = self._check_transitions(state1, state2, alphabet)
        return (r1 or ignore_initial_weight) and r2 and r3

    @abstractmethod
    def equivalent_values(self, value1, value2):
        raise NotImplementedError

    def _check_transitions(self, state1, state2, alphabet):
        for symbol in alphabet.symbols:
            transitions1 = list(state1.transitions_set[symbol])
            transitions2 = list(state2.transitions_set[symbol])
            if len(transitions1) > 1 or len(transitions2) > 1:
                # TODO custom exception
                raise Exception(
                    "Eq method supported only for Deterministic Weighted FA")
            transition1 = transitions1[0]
            transition2 = transitions2[0]
            if not self.equivalent_values(transition1.weight, transition2.weight):
                return False
        return True

    def _process_equivalence_iteration_with(self, symbol, pairs_to_visit, visited_pairs, sequence_for_pairs):
        pair = pairs_to_visit[0]
        self_next_state = min(pair[0].next_states_for(symbol))
        other_next_state = min(pair[1].next_states_for(symbol))
        next_pair = (self_next_state, other_next_state)
        next_pair_in_pairs_to_visit = any(self._pair_equivalent_by_name(
            elem, next_pair) for elem in pairs_to_visit)
        next_pair_in_visited_pairs = any(self._pair_equivalent_by_name(
            elem, next_pair) for elem in visited_pairs)
        if not next_pair_in_pairs_to_visit and not next_pair_in_visited_pairs:
            sequence_for_pairs[next_pair] = sequence_for_pairs[pair] + symbol
            pairs_to_visit.append(next_pair)



from unittest import TestLoader, TestSuite, TextTestRunner
import os
import sys

from pythautomata.tests.test_minimizer import TestMinimizer
from pythautomata.tests.test_automata_convertor import TestAutomataConvertor
from pythautomata.tests.test_moore_machines_last_symbol_queries import TestMooreMachinesLastSymbolQueries
from pythautomata.tests.test_pdfa_generator import TestPDFAGenerator
from pythautomata.tests.test_pdfa_serialization import TestPDFASerialization
from pythautomata.tests.test_dfa_serialization import TestDFASerialization
from pythautomata.tests.test_dfa_generators import TestDFAGenerators
from pythautomata.tests.test_automata_comparison import TestAutomataComparison
from pythautomata.tests.test_sequence import TestSequence
from pythautomata.tests.test_dfa_operations import TestDFAOperations
from pythautomata.tests.test_automata_definitions import TestAutomataDefinitions
from pythautomata.tests.test_regex_generator import TestRegexGenerator
from pythautomata.tests.test_pdfa_metrics import TestPDFAMetrics
from pythautomata.tests.test_simple_DFA_generator import TestSimpleDFAGenerator
from pythautomata.tests.test_pdfa_operations import TestPDFAOperations
from pythautomata.tests.test_pdfa_last_token_queries import TestPDFALastTokenQueries
from pythautomata.tests.test_dfa_to_moore_machine import TestDFAToMooreMachine
from pythautomata.tests.test_dfa_exporting import TestDFAExporting
from pythautomata.tests.test_dfa_loading import TestDFALoading
from pythautomata.tests.test_probabilistic_filter_model import TestProbabilisticFilterModel
from pythautomata.tests.test_mm_exporting import TestMMExporting
from pythautomata.tests.test_mm_loading import TestMMLoading
from pythautomata.tests.test_mealy_exporting import TestMealyExporting
from pythautomata.tests.test_wfa_exporting import TestWFAExporting
from pythautomata.tests.test_mealy_loading import TestMealyLoading
from pythautomata.tests.test_composed_boolean_model import TestComposedBooleanModel
from pythautomata.tests.test_moore_generators import TestMooreGenerators
from pythautomata.tests.test_composed_probabilistic_model import TestComposedProbabilisticModel
from pythautomata.tests.test_sequence_generator import TestSequenceGenerator

def get_all_tests():
    return [TestDFALoading, TestDFAExporting, TestSimpleDFAGenerator, TestMinimizer,
                TestAutomataConvertor, TestDFAGenerators, TestAutomataComparison, TestSequence,
                TestDFAOperations, TestAutomataDefinitions, TestPDFAGenerator, TestRegexGenerator,
                TestPDFASerialization, TestDFASerialization, TestPDFAMetrics, TestPDFAOperations,
                TestPDFALastTokenQueries, TestMooreMachinesLastSymbolQueries, TestDFAToMooreMachine,
                TestMMLoading, TestMMExporting, TestProbabilisticFilterModel, TestMealyExporting, TestWFAExporting,
                TestMealyLoading, TestComposedBooleanModel, TestMooreGenerators, TestComposedProbabilisticModel,TestSequenceGenerator]


def run():
    test_classes_to_run = get_all_tests()
    loader = TestLoader()
    suites_list = []
    for test_class in test_classes_to_run:
        suite = loader.loadTestsFromTestCase(test_class)
        suites_list.append(suite)
    meta_suite = TestSuite(suites_list)
    TextTestRunner().run(meta_suite)


if __name__ == '__main__':
    run()



from unittest import TestCase
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence
from pythautomata.utilities.uniform_length_sequence_generator import UniformLengthSequenceGenerator
from pythautomata.utilities.uniform_word_sequence_generator import UniformWordSequenceGenerator
from pythautomata.utilities.guiding_wfa_sequence_generator import GuidingWDFASequenceGenerator
from pythautomata.automata_definitions.weighted_tomitas_grammars import WeightedTomitasGrammars
abcAlphabet = Alphabet(frozenset(
    [SymbolStr('a'), SymbolStr('b'), SymbolStr('c')]))


class TestSequenceGenerator(TestCase):

    def test_sequence_generator(self):
        guide = WeightedTomitasGrammars().get_automaton_1()
        uniformLengthGenerator = UniformLengthSequenceGenerator(abcAlphabet, 10, 30, 0)
        uniformWordGenerator = UniformWordSequenceGenerator(abcAlphabet, 10, 30, 0)
        gudingWDFAGenerator = GuidingWDFASequenceGenerator(guide, 10, 30, 0.3)
        generators = [uniformLengthGenerator, uniformWordGenerator, gudingWDFAGenerator]
        for generator in generators:
            generator.generate_words(100)


from unittest import TestCase

from pythautomata.automata_definitions.weighted_tomitas_grammars import WeightedTomitasGrammars
from pythautomata.other_models.composed_probabilistic_model import ComposedProbabilisticModel
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.symbol import SymbolStr


class TestComposedProbabilisticModel(TestCase):

    def test_instantiate_composed_probabilistic_model_1(self):
        automata1 = WeightedTomitasGrammars.get_automaton_1()
        automata2 = WeightedTomitasGrammars.get_automaton_2()
        model = ComposedProbabilisticModel(
            [automata1, automata2], automata1.alphabet, lambda x, y: (x + y)/2)
        a = model.sequence_probability(
            Sequence([SymbolStr('0'), SymbolStr('1')]))
        b = model.last_token_probability(
            Sequence([SymbolStr('0'), SymbolStr('1')]))
        self.assertTrue(True)

    def test_instantiate_composed_probabilistic_model_exception(self):
        automata1 = WeightedTomitasGrammars.get_automaton_1()
        automata2 = WeightedTomitasGrammars.get_automaton_2()
        with self.assertRaises(TypeError):
            ComposedProbabilisticModel(
                [automata1, automata2], automata1.alphabet, lambda x, y, z: (x and not y or z))



from pythautomata.abstract.model_exporting_strategy import ModelExportingStrategy


class EncodedFileExportingStrategy(ModelExportingStrategy):

    def export(self, model, path=None):
        path = self.get_path_for(path, model)
        path = str(path)
        fileObject = open(path + ".ef", "w+", encoding="utf-8")

        fileObject.write("Alphabet: [")
        fileObject.write(
            ", ".join(map(lambda symbol: f"'{str(symbol)}'", sorted(model.alphabet.symbols))))

        fileObject.write("]\n\nStates:")
        sortedStates = sorted(model.states, key=lambda state: state.name)
        for state in sortedStates:
            fileObject.write(f"\n\t\"{state.name}\"")
            if state in model.initial_states:
                fileObject.write(" (Initial)")
            if state.is_final:
                fileObject.write(" (Final)")

        fileObject.write("\n\nTransitions:\n\n")
        fileObject.write("\n".join(map(self._get_all_transitions_lines_for,
                                       sortedStates)))
        fileObject.close()

    def _get_all_transitions_lines_for(self, state):
        if not state.transitions:
            return ""
        sortedTransitions = sorted(state.transitions.items())
        transitionsLines = "\n\t".join(
            map(self._get_one_symbol_transitions_lines_for,
                sortedTransitions))
        return f"\"{state.name}\":\n\t{transitionsLines}"

    def _get_one_symbol_transitions_lines_for(self, item):
        substring = ", ".join(
            map(lambda next_state: f"\"{next_state.name}\"", item[1]))
        return f"'{str(item[0])}' -> [{substring}]"



from unittest import TestCase

from pythautomata.automata_definitions.other_automata import OtherAutomata
from pythautomata.automata_definitions.sample_mealy_machines import SampleMealyMachines
from pythautomata.utilities.automata_converter import AutomataConverter
from pythautomata.automata_definitions.sample_moore_machines import SampleMooreMachines
from pythautomata.model_comparators.mealy_machine_comparison_strategy import MealyMachineComparisonStrategy

class TestAutomataConvertor(TestCase):

    def test_conversion_1(self):
        automaton = OtherAutomata.get_nfa_1()
        automata_converted_to_dfa = AutomataConverter.convert_nfa_to_dfa(
            automaton)
        expected_converted_dfa = OtherAutomata.get_dfa_1()
        self.assertEqual(automata_converted_to_dfa, expected_converted_dfa)

    def test_conversion_2(self):
        automaton = OtherAutomata.get_nfa_2()
        automata_converted_to_dfa = AutomataConverter.convert_nfa_to_dfa(
            automaton)
        expected_converted_dfa = OtherAutomata.get_dfa_2()
        self.assertEqual(automata_converted_to_dfa, expected_converted_dfa)

    def test_conversion_3(self):
        automaton = OtherAutomata.get_nfa_3()
        automata_converted_to_dfa = AutomataConverter.convert_nfa_to_dfa(
            automaton)
        expected_converted_dfa = OtherAutomata.get_dfa_3()
        self.assertEqual(automata_converted_to_dfa, expected_converted_dfa)

    def test_conversion_4(self):
        automaton = OtherAutomata.get_nfa_4()
        automata_converted_to_dfa = AutomataConverter.convert_nfa_to_dfa(
            automaton)
        expected_converted_dfa = OtherAutomata.get_dfa_4()
        self.assertEqual(automata_converted_to_dfa, expected_converted_dfa)

    def test_conversion_5(self):
        automaton = OtherAutomata.get_nfa_5()
        automata_converted_to_dfa = AutomataConverter.convert_nfa_to_dfa(
            automaton)
        expected_converted_dfa = OtherAutomata.get_dfa_5()
        self.assertEqual(automata_converted_to_dfa, expected_converted_dfa)

    def test_conversion_moore_to_mealy(self):
        machine = SampleMooreMachines.get_tomitas_automaton_1()
        converted_mealy_machine = AutomataConverter.convert_moore_machine_to_mealy_machine(
            machine)
        expected_converted_mealy_machine = SampleMealyMachines.get_tomitas_automaton_1()
        comparator = MealyMachineComparisonStrategy()
        self.assertTrue(comparator.are_equivalent(converted_mealy_machine, expected_converted_mealy_machine))
        
    def test_conversion_moore_to_minimal_mealy(self):
        machine = SampleMooreMachines.get_minimizable_for_mealy_moore_1()
        
        converted_mealy_machine = AutomataConverter.\
            convert_moore_machine_to_minimal_mealy_machine(machine)

        expected_converted_mealy_machine = SampleMealyMachines.get_minimized_mealy_1()

        comparator = MealyMachineComparisonStrategy()
        self.assertTrue(comparator.are_equivalent(converted_mealy_machine, expected_converted_mealy_machine))
        

    def test_conversion_moore_to_minimal_mealy_2(self):
        machine = SampleMooreMachines.get_minimizable_for_mealy_moore_2()

        converted_mealy_machine = AutomataConverter.\
                convert_moore_machine_to_minimal_mealy_machine(machine)
        
        expected_converted_mealy_machine = SampleMealyMachines.get_minimized_mealy_2()

        comparator = MealyMachineComparisonStrategy()
        self.assertTrue(comparator.are_equivalent(converted_mealy_machine, expected_converted_mealy_machine))




from typing import Any
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.symbol import Symbol
from pythautomata.abstract.probabilistic_model import ProbabilisticModel
import numpy as np


class ComposedProbabilisticModel(ProbabilisticModel):
    def __init__(self, models: list, alphabet: Alphabet, reduction_function=lambda *args: np.mean(args), verbose: bool = False):
        """
        Probabilistic model consisting of composing a series of probabilistic models

        Args:
            models (list): A list of models to process
            alphabet (Alphabet): The alphabet of the models
            reduction_function (_type_, optional): A function for reducing the output of the models. Defaults to lambda*args:np.mean(args).            
        """
        # init object
        self._models = models
        self._alphabet = alphabet
        self._reduction_function = reduction_function
        self._terminal_symbol = models[0].terminal_symbol

        self._verify_reduction_function(reduction_function, models)
        self._verify_models(models)

    @property
    def terminal_symbol(self) -> Symbol:
        return self._terminal_symbol

    @property
    def alphabet(self) -> Alphabet:
        return self._alphabet

    @property
    def name(self) -> str:
        name = "Composed_model:"
        for model in self._models:
            name = name + " - " + model.name
        return name

    def _verify_reduction_function(self, fun, models: list):
        try:
            args = map(lambda x: True, models)
            fun(*args)
        except TypeError:
            print("Reduction function should accept as many parameters as len(models)")
            raise

    def _verify_models(self, models: Any):
        all_verify = all(map(lambda x: hasattr(
            x, "last_token_probability"), models))
        assert all_verify, "Make sure all the models are of the class ProbabilisticModel"
        for model in models:
            assert model.terminal_symbol == self.terminal_symbol, "All models should share the terminal symbol"

    def sequence_probability(self, sequence: Sequence) -> float:
        return self._reduction_function(*map(lambda x: x.sequence_probability(sequence), self._models))

    def log_sequence_probability(self, sequence: Sequence) -> float:
        raise NotImplementedError

    def last_token_probability(self, sequence: Sequence) -> float:
        return self._reduction_function(*map(lambda x: x.last_token_probability(sequence), self._models))



from pythautomata.automata.deterministic_finite_automaton import \
    DeterministicFiniteAutomaton
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.state import State
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.model_comparators.dfa_comparison_strategy import \
    DFAComparisonStrategy as DFAComparator

abcAlphabet = Alphabet(frozenset(
    (SymbolStr('a'), SymbolStr('b'), SymbolStr('c'))))
abcdAlphabet = Alphabet(frozenset(
    (SymbolStr('a'), SymbolStr('b'), SymbolStr('c'), SymbolStr('d'))))


class OmlinGilesAutomata:
    """
    Class containing automata from paper:
        Christian W. Omlin and C. Lee Giles. 1996. 
        Constructing deterministic finite-state automata in recurrent neural networks. 
        J. ACM 43, 6 (Nov. 1996), 937–972. DOI:https://doi.org/10.1145/235809.235811

    Methods
    -------
    get_all_automata: list(DeterministicFiniteAutomaton)
        returns a list containing all automata defined in this class

    get_a_automaton: DeterministicFiniteAutomaton
        returns the automaton A from the paper 

    get_b_automaton: DeterministicFiniteAutomaton
        returns the automaton B from the paper

    get_az_automaton: DeterministicFiniteAutomaton
        returns the automaton A from the paper with stuttering symbol (d)

    get_bz_automaton: DeterministicFiniteAutomaton
        returns the automaton A from the paper with stuttering symbol (d)
    """

    @staticmethod
    def get_all_automata():
        """
        method returning a list of all automata of the class

        Returns
        -------
        list(DeterministicFiniteAutomaton)
            all automata defined in the class

        """
        return [
            OmlinGilesAutomata.get_a_automaton(),
            OmlinGilesAutomata.get_az_automaton(),
            OmlinGilesAutomata.get_b_automaton(),
            OmlinGilesAutomata.get_bz_automaton()
        ]

    @staticmethod
    def get_a_automaton():
        """
        method with specification of the automaton A from the paper

        Returns
        -------
        DeterministicFiniteAutomaton
            automaton A from the paper

        """
        a = abcAlphabet['a']
        b = abcAlphabet['b']
        c = abcAlphabet['c']
        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2", True)
        state0.add_transition(a, state0)
        state0.add_transition(b, state1)
        state0.add_transition(c, state1)
        state1.add_transition(a, state1)
        state1.add_transition(b, state0)
        state1.add_transition(c, state2)
        state2.add_transition(a, state1)
        state2.add_transition(b, state0)
        state2.add_transition(c, state2)
        comparator = DFAComparator()
        return DeterministicFiniteAutomaton(abcAlphabet, state0,
                                            set([state0, state1, state2]), comparator, "a automaton",)

    @staticmethod
    def get_b_automaton():
        """
        method with specification of the automaton B from the paper

        Returns
        -------
        DeterministicFiniteAutomaton
            automaton B from the paper
        """
        a = abcAlphabet['a']
        b = abcAlphabet['b']
        c = abcAlphabet['c']
        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2")
        state3 = State("State 3")
        state4 = State("State 4")
        state5 = State("State 5", True)
        state0.add_transition(a, state0)
        state0.add_transition(b, state3)
        state0.add_transition(c, state3)
        state1.add_transition(a, state3)
        state1.add_transition(b, state1)
        state1.add_transition(c, state4)
        state2.add_transition(a, state4)
        state2.add_transition(b, state5)
        state2.add_transition(c, state2)
        state3.add_transition(a, state0)
        state3.add_transition(b, state1)
        state3.add_transition(c, state1)
        state4.add_transition(a, state1)
        state4.add_transition(b, state1)
        state4.add_transition(c, state2)
        state5.add_transition(a, state0)
        state5.add_transition(b, state0)
        state5.add_transition(c, state2)
        comparator = DFAComparator()
        return DeterministicFiniteAutomaton(abcAlphabet, state0,
                                            set([state0, state1, state2, state3,
                                                 state4, state5]), comparator, "b automaton")

    @staticmethod
    def get_az_automaton():
        """
        method with specification of the automaton A from the paper, with a stuttering symbol

        Returns
        -------
        DeterministicFiniteAutomaton
            automaton A from the paper, with a stuttering symbol
        """
        a = abcdAlphabet['a']
        b = abcdAlphabet['b']
        c = abcdAlphabet['c']
        d = abcdAlphabet['d']
        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2", True)
        state0.add_transition(a, state0)
        state0.add_transition(d, state0)
        state0.add_transition(b, state1)
        state0.add_transition(c, state1)
        state1.add_transition(a, state1)
        state1.add_transition(b, state0)
        state1.add_transition(c, state2)
        state1.add_transition(d, state1)
        state2.add_transition(a, state1)
        state2.add_transition(b, state0)
        state2.add_transition(c, state2)
        state2.add_transition(d, state2)
        comparator = DFAComparator()
        return DeterministicFiniteAutomaton(abcdAlphabet, state0,
                                            set([state0, state1, state2]), comparator, "az automaton")

    @staticmethod
    def get_bz_automaton():
        """
        method with specification of the automaton B from the paper, with a stuttering symbol

        Returns
        -------
        DeterministicFiniteAutomaton
            automaton B from the paper, with a stuttering symbol
        """
        a = abcdAlphabet['a']
        b = abcdAlphabet['b']
        c = abcdAlphabet['c']
        d = abcdAlphabet['d']
        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2")
        state3 = State("State 3")
        state4 = State("State 4")
        state5 = State("State 5", True)
        state0.add_transition(a, state0)
        state0.add_transition(b, state3)
        state0.add_transition(c, state3)
        state0.add_transition(d, state0)
        state1.add_transition(a, state3)
        state1.add_transition(b, state1)
        state1.add_transition(c, state4)
        state1.add_transition(d, state1)
        state2.add_transition(a, state4)
        state2.add_transition(b, state5)
        state2.add_transition(c, state2)
        state2.add_transition(d, state2)
        state3.add_transition(a, state0)
        state3.add_transition(b, state1)
        state3.add_transition(c, state1)
        state3.add_transition(d, state3)
        state4.add_transition(a, state1)
        state4.add_transition(b, state1)
        state4.add_transition(c, state2)
        state4.add_transition(d, state4)
        state5.add_transition(a, state0)
        state5.add_transition(b, state0)
        state5.add_transition(c, state2)
        state5.add_transition(d, state5)
        comparator = DFAComparator()
        return DeterministicFiniteAutomaton(abcdAlphabet, state0,
                                            set([state0, state1, state2, state3,
                                                 state4, state5]), comparator, "bz automaton")



from pythautomata.automata.mealy_machine import MealyMachine
from pythautomata.base_types.sequence import Sequence


class MealyMachineComparisonStrategy:
    def are_equivalent(self, machine1: MealyMachine, machine2: MealyMachine):
        return self.get_counterexample_between(machine1, machine2) is None

    def get_counterexample_between(self, machine1: MealyMachine, machine2: MealyMachine):
        if machine1._alphabet != machine2._alphabet:
            raise ValueError("Input alphabets are not equivalent.")
        if machine1.output_alphabet != machine2.output_alphabet:
            raise ValueError("Output alphabets are not equivalent.")

        initial_pair = (machine1.initial_state, machine2.initial_state)
        pairs_to_visit = [initial_pair]
        sequence_for_pairs = {initial_pair: Sequence()}
        visited_pairs = set()

        pair = pairs_to_visit[0]
        for symbol in machine1._alphabet.symbols:
            if pair[0].outputs[symbol] != pair[1].outputs[symbol]:
                return sequence_for_pairs[pair]
            self_next_state = min(pair[0].next_states_for(symbol))
            other_next_state = min(pair[1].next_states_for(symbol))
            next_pair = (self_next_state, other_next_state)
            if next_pair not in pairs_to_visit and next_pair not in visited_pairs:
                sequence_for_pairs[next_pair] = sequence_for_pairs[pair] + symbol
                pairs_to_visit.append(next_pair)
        return None



from pythautomata.model_comparators.wfa_comparison_strategy import WFAComparator
from pythautomata.utilities import pdfa_utils


class WFAToleranceComparator(WFAComparator):
    """
    Class containing a WFA Comparator Strategy based on next symbol distributions and a tolerance parameter.

    Methods
    -------   
    are_equivalent: bool
        returns true iif wfa1 is equivalent to wfa2 according to a given tolerance

    get_counterexample_between: Sequence
        returns a Sequence where the next token weights differ (or its difference is greater than the tolerance)
    """

    def __init__(self, tolerance: float = 0) -> None:
        super().__init__()
        self.tolerance = tolerance

    def equivalent_output(self, observation1, observation2) -> bool:
        return pdfa_utils.are_within_tolerance_limit(observation1, observation2, self.tolerance)

    def equivalent_values(self, value1, value2):
        return abs(value1 - value2) <= self.tolerance

    def next_tokens_equivalent_output(self, observation1, observation2) -> bool:
        return self.equivalent_output(observation1, observation2)



from unittest import TestCase

from pythautomata.automata_definitions.sample_moore_machines import SampleMooreMachines
from pythautomata.model_exporters.dot_exporters.moore_dot_exporting_strategy import MooreDotExportingStrategy
from pythautomata.model_exporters.standard_exporters.moore_standard_dot_exporting_strategy import MooreStandardDotExportingStrategy


class TestMMExporting(TestCase):

    def test_1(self):
        mm = SampleMooreMachines.get_3_states_automaton()
        MooreDotExportingStrategy().export(mm, "./output_models/tests")
        MooreStandardDotExportingStrategy().export(mm, "./output_models/tests")

    def test_2(self):
        mm = SampleMooreMachines.get_tomitas_automaton_1()
        MooreDotExportingStrategy().export(mm, "./output_models/tests")
        MooreStandardDotExportingStrategy().export(mm, "./output_models/tests")

    def test_3(self):
        mm = SampleMooreMachines.get_tomitas_automaton_2()
        MooreDotExportingStrategy().export(mm, "./output_models/tests")
        MooreStandardDotExportingStrategy().export(mm, "./output_models/tests")



from pythautomata.automata.mealy_machine import MealyMachine
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.mealy_state import MealyState
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.model_comparators.mealy_machine_comparison_strategy import MealyMachineComparisonStrategy


abAlphabet = Alphabet(frozenset((SymbolStr('a'), SymbolStr('b'))))
abcAlphabet = Alphabet(
    frozenset((SymbolStr('a'), SymbolStr('b'), SymbolStr('c'))))


class SampleMealyMachines:
    """
    Class containing sample Moore Machines

    Methods
    -------
    get_all_mealy_machines()
        Returns a list containing all the sample Mealy Machines
    get_3_states_mealy_machine()
        Returns a Mealy Machine with 3 states
    get_tomitas_automaton_1()
        Returns a Mealy Machine that accepts the language of Tomita's automaton 1
    """

    @staticmethod
    def get_all_mealy_machines():
        return [
            SampleMealyMachines.get_3_states_mealy_machine(),
            SampleMealyMachines.get_tomitas_automaton_1(),
        ]

    @staticmethod
    def get_3_states_mealy_machine():
        a = abAlphabet['a']
        b = abAlphabet['b']

        state0 = MealyState("State 0")
        state1 = MealyState("State 1")
        state2 = MealyState("State 2")

        state0.add_transition(a, state1, abcAlphabet['a'])
        state0.add_transition(b, state2, abcAlphabet['b'])
        state1.add_transition(a, state2, abcAlphabet['c'])
        state1.add_transition(b, state0, abcAlphabet['a'])
        state2.add_transition(a, state0, abcAlphabet['b'])
        state2.add_transition(b, state1, abcAlphabet['c'])

        return MealyMachine(abAlphabet, abcAlphabet, state0,
                            set([state0, state1, state2]), MealyMachineComparisonStrategy(), "3 States Mealy Machine")

    @staticmethod
    def get_tomitas_automaton_1():
        alphabet = Alphabet(frozenset((SymbolStr('False'), SymbolStr('True'))))

        binaryAlphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
        zero = binaryAlphabet['0']
        one = binaryAlphabet['1']

        stateA = MealyState("State A")
        stateB = MealyState("State B")
        stateA.add_transition(one, stateA, alphabet['True'])
        stateA.add_transition(zero, stateB, alphabet['False'])
        stateB.add_transition(one, stateB, alphabet['False'])
        stateB.add_transition(zero, stateB, alphabet['False'])

        hole_state = MealyState(name="hole")

        return MealyMachine(binaryAlphabet, alphabet, stateA,
                            set([stateA, stateB]),
                            MealyMachineComparisonStrategy(),
                            name="Mealy machine implementation of Tomita's grammar 2 automaton",
                            hole=hole_state)
    
    @staticmethod
    def get_minimized_mealy_1():
        alphabet = Alphabet(frozenset((SymbolStr('N'), SymbolStr('Y'))))

        binaryAlphabet = Alphabet(frozenset((SymbolStr('H'), SymbolStr('T'))))
        heads = binaryAlphabet['H']
        tails = binaryAlphabet['T']

        state0 = MealyState("q0")
        state1 = MealyState("q1")
        state0.add_transition(heads, state0, alphabet['N'])
        state0.add_transition(tails, state1, alphabet['N'])
        state1.add_transition(heads, state0, alphabet['N'])
        state1.add_transition(tails, state1, alphabet['Y'])

        hole_state = MealyState(name="hole")

        return MealyMachine(binaryAlphabet, alphabet, state0,
                            set([state0, state1]),
                            MealyMachineComparisonStrategy(),
                            name="Minimized mealy machine example",
                            hole=hole_state)

    @staticmethod
    def get_minimized_mealy_2():
        input_alphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
        output_alphabet = Alphabet(frozenset((SymbolStr('N'), SymbolStr('Y'))))

        state0 = MealyState("q0")
        state1 = MealyState("q1")
        state2 = MealyState("q2")
        state3 = MealyState("q3")

        state0.add_transition(input_alphabet['0'], state1, output_alphabet['N'])
        state0.add_transition(input_alphabet['1'], state0, output_alphabet['N'])
        state1.add_transition(input_alphabet['0'], state1, output_alphabet['N'])
        state1.add_transition(input_alphabet['1'], state2, output_alphabet['N'])
        state2.add_transition(input_alphabet['0'], state3, output_alphabet['N'])
        state2.add_transition(input_alphabet['1'], state0, output_alphabet['N'])
        state3.add_transition(input_alphabet['0'], state1, output_alphabet['N'])
        state3.add_transition(input_alphabet['1'], state2, output_alphabet['Y'])

        hole_state = MealyState(name="hole")

        return MealyMachine(input_alphabet, output_alphabet, state0,
                            set([state0, state1, state2, state3]),
                            MealyMachineComparisonStrategy(),
                            name="Minimized mealy machine example 2",
                            hole=hole_state)



from random import randint, seed
import random

from pythautomata.automata.mealy_machine import MealyMachine
from pythautomata.base_types.sequence import Sequence


class RandomWalkMealyMachineComparisonStrategy:

    def __init__(self, number_steps: int, reset_probability: float, random_seed: int = 21, ) -> None:
        self._number_steps = number_steps
        self._reset_probability = reset_probability
        self._seed = random_seed
        seed(self._seed)

    def are_equivalent(self, machine1: MealyMachine, machine2: MealyMachine) -> bool:
        return self.get_counterexample_between(machine1, machine2) is None

    def get_counterexample_between(self, machine1: MealyMachine, machine2: MealyMachine):
        if machine1._alphabet != machine2._alphabet:
            raise ValueError("Input alphabets are not equivalent.")
        if machine1._output_alphabet != machine2._output_alphabet:
            raise ValueError("Output alphabets are not equivalent.")

        steps = 0
        counter_example = Sequence()
        machine1.reset()
        machine2.reset()
        if not self.equivalent_output(
            machine1.last_symbol(counter_example), 
            machine2.last_symbol(counter_example)):
            return counter_example

        symbols = list(machine1._alphabet.symbols)
        symbols.sort()

        while steps < self._number_steps:
            if random.random() <= self._reset_probability:
                counter_example = Sequence()
                machine1.reset()
                machine2.reset()
            pos = randint(0, len(symbols) - 1)
            counter_example = counter_example.append(symbols[pos])
            steps += 1
            if not self.equivalent_output(machine1.step(symbols[pos]), machine2.step(symbols[pos])):
                return counter_example

        return None

    def equivalent_output(self, observation1, observation2) -> bool:
        return observation1 == observation2



from typing import Pattern
import uuid
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence
from pythautomata.abstract.boolean_model import BooleanModel


class RegularExpression(BooleanModel):

    def __init__(self, alphabet: Alphabet, pattern: Pattern, name: str = None):
        self._alphabet = alphabet
        self._pattern = pattern
        self._name = 'REGEX - ' + str(uuid.uuid4().hex) if name is None else name

    @property
    def alphabet(self) -> Alphabet:
        return self._alphabet

    @property
    def name(self):
        return self._name

    def accepts(self, sequence: Sequence) -> bool:
        return bool(self._pattern.match("".join(map(str, sequence.value))))



from unittest import TestCase

from pythautomata.automata_definitions.sample_mealy_machines import SampleMealyMachines
from pythautomata.model_exporters.standard_exporters.mealy_standard_dot_exporting_strategy import MealyStandardDotExportingStrategy
from pythautomata.utilities.standard_dot_notation_mealy_importer import StandardDotNotationMealyImporter


class TestMealyLoading(TestCase):

    def test_1(self):
        for mealy_machine in SampleMealyMachines.get_all_mealy_machines():
            name = mealy_machine._name
            mealy_machine._name = name + "_standard"

            MealyStandardDotExportingStrategy().export(
                mealy_machine, "./output_models/tests")

            path = './output_models/tests'+'/'+mealy_machine._name+'.dot'
            loaded_mealy = StandardDotNotationMealyImporter().import_automata(path)

            self.assertEqual(mealy_machine, loaded_mealy)



from typing import Any
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.abstract.boolean_model import BooleanModel


class ComposedBooleanModel(BooleanModel):
    def __init__(self, models: list, alphabet: Alphabet, reduction_function=lambda *args: all(args), verbose: bool = False):
        """
        Boolean model consisting of composing a series of different boolean models

        Args:
            models (list): A list of models to process
            alphabet (Alphabet): The alphabet of the models
            reduction_function (_type_, optional): A function for reducing the output of the models. Defaults to lambda*args:all(args).            
        """
        self._verify_reduction_function(reduction_function, models)
        self._verify_models(models)

        # init object
        self._models = models
        self._alphabet = alphabet
        self._reduction_function = reduction_function

    @property
    def alphabet(self) -> Alphabet:
        return self._alphabet

    @property
    def name(self) -> str:
        name = "Composed_model:"
        for model in self._models:
            name = name + " - " + model.name
        return name

    def accepts(self, sequence: Sequence) -> bool:
        accepts = list(map(lambda x: x.accepts(sequence), self._models))
        result = self._reduction_function(*accepts)
        return result

    def accepts_batch(self, sequences) -> list[bool]:
        for model in self._models:
            assert (hasattr(model, 'accepts_batch'))
        results = list(map(lambda x: x.accepts_batch(sequences), self._models))
        intermediate_results = zip(*results)
        return list(map(lambda x: self._reduction_function(*x), intermediate_results))

    def _verify_reduction_function(self, fun, models: list):
        try:
            args = map(lambda x: True, models)
            fun(*args)
        except TypeError:
            print("Reduction function should accept as many parameters as len(models)")
            raise

    def _verify_models(self, models: Any):
        all_verify = all(map(lambda x: hasattr(x, "accepts"), models))
        assert all_verify, "Make sure all the models are of the class BooleanModel"



import numpy as np
from abc import ABC, abstractmethod


class ProbabilityPartitioner(ABC):

    def __init__(self):
        self._partitions_cache = dict()

    def get_partition(self, probability_vector):
        if tuple(probability_vector) in self._partitions_cache:
            return self._partitions_cache[tuple(probability_vector)]
        else:
            partition = self._get_partition(probability_vector)
            self._partitions_cache[tuple(probability_vector)] = partition
            return partition

    @abstractmethod
    def _get_partition(self, probability_vector):
        raise NotImplementedError

    def are_in_same_partition(self, probability_vector1, probability_vector2):
        assert (len(probability_vector1) == len(probability_vector2))
        partition1 = self.get_partition(probability_vector1)
        partition2 = self.get_partition(probability_vector2)
        return np.all(np.equal(partition1, partition2))


class QuantizationProbabilityPartitioner(ProbabilityPartitioner):

    def __init__(self, number_of_partitions) -> None:
        super().__init__()
        self._partitions = number_of_partitions

    def _get_interval(self, value):
        assert (value >= 0 and value <= 1)
        limits = np.linspace(0, 1, self._partitions+1)
        if value == 1:
            return self._partitions-1
        positions = list(range(len(limits)-1))
        mid_element = int(len(limits)/2)
        while len(positions) > 1:
            if value >= limits[mid_element]:
                positions = positions[int(len(positions)/2):]
            else:
                positions = positions[:int(len(positions)/2)]
            mid_element = positions[int(len(positions)/2)]
        assert (len(positions) == 1)
        return positions[0]

    def _get_partition(self, probability_vector):
        return np.fromiter((self._get_interval(xi) for xi in probability_vector), dtype=int)


class TopKProbabilityPartitioner(ProbabilityPartitioner):

    def __init__(self, k) -> None:
        self.k = k
        super().__init__()

    def _get_partition(self, probability_vector):
        order = (np.array(probability_vector)*-1).argsort()
        ranks = order.argsort()
        ranks[ranks >= self.k] = -1
        ranks[ranks != -1] = 1
        return ranks


class RankingPartitioner(ProbabilityPartitioner):

    def __init__(self, k) -> None:
        self.k = k
        super().__init__()

    def _get_partition(self, probability_vector):
        order = (np.array(probability_vector)*-1).argsort()
        ranks = order.argsort()
        ranks[ranks >= self.k] = -1
        return ranks



from genericpath import isdir
from os import makedirs
from pathlib import Path
from pythautomata.abstract.model_exporting_strategy import ModelExportingStrategy


class ImageExportingStrategy:

    '''
    Exports image of an automata generated by any exporting strategy based on Graphviz. 
    Works for DFA, Moore, Mealy and WFA

    ----------
    Attributes
    ----------
    exporter: ModelExportingStrategy
        Exporting strategy that will be used to generate the image.
    format: str
        Format of the image to be generated. It can be any format supported by Graphviz. (e.g., png, svg, pdf, etc.) 
    '''

    def __init__(self, exporting_strategy, format: str):
        self.exporter = exporting_strategy
        self.format = format

    def export(self, model, path):
        graph = self.exporter.create_graph(model)
        path = self.get_path_for(path, model)
        graph.render(path, cleanup=True, format=self.format)

    def get_path_for(self, path, model):
        if path is None:
            name = model.name
            path = "output_models/" + \
                ("" if name is None else f"{name}")
        if not isdir(path):
            makedirs(path)
        return Path(path, model.name)



from typing import Optional
import random

from pythautomata.abstract.finite_automaton import FiniteAutomataComparator
from pythautomata.base_types.sequence import Sequence
from random import seed
from random import randint


class RandomWalkComparisonStrategy(FiniteAutomataComparator):

    def __init__(self, number_steps: int, reset_probability: float, random_seed: int = 21, ) -> None:
        super().__init__()
        self._number_steps = number_steps
        self._reset_probability = reset_probability
        self._seed = random_seed
        seed(self._seed)

    def are_equivalent(self, automaton1, automaton2) -> bool:
        return self.get_counterexample_between(automaton1, automaton2) is None

    # TODO: Change types to DeterministicFiniteAutomaton
    def get_counterexample_between(self, dfa1, dfa2) -> Optional[Sequence]:
        if dfa1.alphabet != dfa2.alphabet:
            raise ValueError("Alphabets are not equivalent.")

        steps = 0
        counter_example = Sequence()
        dfa1.reset()
        dfa2.reset()
        if not self.equivalent_output(
                dfa1.accepts(counter_example), dfa2.accepts(counter_example)):
            return counter_example

        symbols = list(dfa1.alphabet.symbols)
        symbols.sort()

        while steps < self._number_steps:
            if random.random() <= self._reset_probability:
                counter_example = Sequence()
                dfa1.reset()
                dfa2.reset()
            pos = randint(0, len(symbols) - 1)
            counter_example = counter_example.append(symbols[pos])
            steps += 1
            if not self.equivalent_output(dfa1.step(symbols[pos]), dfa2.step(symbols[pos])):
                return counter_example
        return None

    def equivalent_output(self, observation1, observation2) -> bool:
        return observation1 == observation2



from pythautomata.base_types.state import State
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.alphabet import Alphabet

from pythautomata.automata.moore_machine_automaton import MooreMachineAutomaton
from pythautomata.base_types.moore_state import MooreState

from pythautomata.model_comparators.moore_machine_comparison_strategy import MooreMachineComparisonStrategy as MooreMachineComparison


abAlphabet = Alphabet(frozenset((SymbolStr('a'), SymbolStr('b'))))
abcAlphabet = Alphabet(
    frozenset((SymbolStr('a'), SymbolStr('b'), SymbolStr('c'))))


class SampleMooreMachines:

    """
    Class containing sample Moore Machines

    Methods
    -------
    get_a_symbol()
        Returns the symbol 'a'
    get_b_symbol()
        Returns the symbol 'b'
    get_c_symbol()
        Returns the symbol 'c'
    get_all_automata()
        Returns a list containing all the sample Moore Machines that are correct
    get_3_states_automaton()
        Returns a Moore Machine with 3 states
    get_3_states_automaton_wrong_alphabet()
        Returns a Moore Machine with 3 states and wrong alphabet. This is used to test the Moore Machine constructor 
        and will raise an exception when used
    get_3_states_automaton_non_deterministic()
        Returns a Moore Machine with 3 states and non deterministic transitions. 
        This is used to test the Moore Machine constructor and will raise an exception when used
    get_tomitas_automaton_1()
        Returns a Moore Machine that accepts the language of Tomita's automaton 1
    get_tomitas_automaton_2()
        Returns a Moore Machine that accepts the language of Tomita's automaton 2

    """
    @staticmethod
    def get_a_symbol():
        return 'a'

    @staticmethod
    def get_b_symbol():
        return 'b'

    @staticmethod
    def get_c_symbol():
        return 'c'

    @staticmethod
    def get_all_automata(calculate_access_strings: bool = False):
        return [
            SampleMooreMachines.get_3_states_automaton(calculate_access_strings),
            SampleMooreMachines.get_tomitas_automaton_1(calculate_access_strings),
            SampleMooreMachines.get_tomitas_automaton_2(calculate_access_strings),
        ]

    @staticmethod
    def get_3_states_automaton(calculate_access_strings = False):
        a = abAlphabet[SampleMooreMachines.get_a_symbol()]
        b = abAlphabet[SampleMooreMachines.get_b_symbol()]

        state0 = MooreState(
            "State 0", abcAlphabet[SampleMooreMachines.get_a_symbol()])
        state1 = MooreState(
            "State 1", abcAlphabet[SampleMooreMachines.get_b_symbol()])
        state2 = MooreState(
            "State 2", abcAlphabet[SampleMooreMachines.get_c_symbol()])

        state0.add_transition(a, state1)
        state0.add_transition(b, state2)
        state1.add_transition(a, state2)
        state1.add_transition(b, state0)
        state2.add_transition(a, state0)
        state2.add_transition(b, state1)

        return MooreMachineAutomaton(abAlphabet, abcAlphabet, state0,
                                     set([state0, state1, state2]), MooreMachineComparison(),
                                     "3 States Moore Machine", 
                                     calculate_access_strings=calculate_access_strings)

    @staticmethod
    def get_3_states_automaton_wrong_alphabet():
        a = abAlphabet[SampleMooreMachines.get_a_symbol()]
        b = abAlphabet[SampleMooreMachines.get_b_symbol()]

        state0 = MooreState(
            "State 0", abcAlphabet[SampleMooreMachines.get_a_symbol()])
        state1 = MooreState(
            "State 1", abcAlphabet[SampleMooreMachines.get_b_symbol()])
        state2 = MooreState(
            "State 2", abcAlphabet[SampleMooreMachines.get_c_symbol()])

        state0.add_transition(a, state1)
        state0.add_transition(b, state2)
        state1.add_transition(a, state2)
        state1.add_transition(b, state0)
        state2.add_transition(a, state0)
        state2.add_transition(b, state1)

        return MooreMachineAutomaton(abAlphabet, abAlphabet, state0,
                                     set([state0, state1, state2]), MooreMachineComparison(), "3 States Moore Machine")

    @staticmethod
    def get_3_states_automaton_non_deterministic():
        a = abAlphabet[SampleMooreMachines.get_a_symbol()]
        b = abAlphabet[SampleMooreMachines.get_b_symbol()]

        state0 = MooreState(
            "State 0", abcAlphabet[SampleMooreMachines.get_a_symbol()])
        state1 = MooreState(
            "State 1", abcAlphabet[SampleMooreMachines.get_b_symbol()])
        state2 = MooreState(
            "State 2", abcAlphabet[SampleMooreMachines.get_c_symbol()])

        state0.add_transition(a, state1)
        state0.add_transition(a, state2)
        state0.add_transition(b, state2)
        state1.add_transition(a, state2)
        state1.add_transition(b, state0)
        state2.add_transition(a, state0)
        state2.add_transition(b, state1)

        return MooreMachineAutomaton(abAlphabet, abAlphabet, state0,
                                     set([state0, state1, state2]),
                                     MooreMachineComparison(), "3 States Moore Machine")

    @staticmethod
    def get_tomitas_automaton_1(calculate_access_strings = False):
        alphabet = Alphabet(frozenset((SymbolStr('False'), SymbolStr('True'))))

        binaryAlphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
        zero = binaryAlphabet['0']
        one = binaryAlphabet['1']

        stateA = MooreState("State A", alphabet['True'])
        stateB = MooreState("State B", alphabet['False'])
        stateA.add_transition(one, stateA)
        stateA.add_transition(zero, stateB)
        stateB.add_transition(one, stateB)
        stateB.add_transition(zero, stateB)

        hole_state = MooreState(name="hole", value=SymbolStr('False'))

        return MooreMachineAutomaton(binaryAlphabet, alphabet, stateA,
                                     set([stateA, stateB]
                                         ), MooreMachineComparison(),
                                     name='2 States Moore Machine', hole=hole_state,
                                     calculate_access_strings=calculate_access_strings)

    @staticmethod
    def get_tomitas_automaton_2(calculate_access_strings = False):
        boolean_alphabet = Alphabet(
            frozenset((SymbolStr('False'), SymbolStr('True'))))

        binaryAlphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
        zero = binaryAlphabet['0']
        one = binaryAlphabet['1']

        stateA = MooreState("State A", boolean_alphabet['True'])
        stateB = MooreState("State B", boolean_alphabet['False'])
        stateC = MooreState("State C", boolean_alphabet['False'])
        stateA.add_transition(one, stateB)
        stateA.add_transition(zero, stateC)
        stateB.add_transition(one, stateC)
        stateB.add_transition(zero, stateA)
        stateC.add_transition(one, stateC)
        stateC.add_transition(zero, stateC)

        hole_state = MooreState("Hole", SymbolStr("False"))

        return MooreMachineAutomaton(binaryAlphabet, boolean_alphabet, stateA,
                                     set([stateA, stateB, stateC]),
                                     MooreMachineComparison(),
                                     name="MMA implementation of Tomita's grammar 2 automaton",
                                     hole=hole_state, calculate_access_strings=calculate_access_strings)
    
    @staticmethod
    def get_minimizable_for_mealy_moore_1():
        input_alphabet = Alphabet(frozenset((SymbolStr('T'), SymbolStr('H'))))
        output_alphabet = Alphabet(frozenset((SymbolStr('N'), SymbolStr('Y'))))

        state0 = MooreState("Q0", output_alphabet['N'])
        state1 = MooreState("Q1", output_alphabet['N'])
        state2 = MooreState("Q2", output_alphabet['Y'])

        state0.add_transition(input_alphabet['T'], state1)
        state0.add_transition(input_alphabet['H'], state0)
        state1.add_transition(input_alphabet['T'], state2)
        state1.add_transition(input_alphabet['H'], state0)
        state2.add_transition(input_alphabet['T'], state2)
        state2.add_transition(input_alphabet['H'], state0)

        hole_state = MooreState("Hole")

        return MooreMachineAutomaton(input_alphabet, output_alphabet, state0,
                                     set([state0, state1, state2]),
                                     MooreMachineComparison(),
                                     name="Sample Moore Machine 1",
                                     hole=hole_state)

    @staticmethod    
    def get_minimizable_for_mealy_moore_2():
        input_alphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
        output_alphabet = Alphabet(frozenset((SymbolStr('N'), SymbolStr('Y'))))

        state0 = MooreState("Q0", output_alphabet['N'])
        state1 = MooreState("Q1", output_alphabet['N'])
        state2 = MooreState("Q2", output_alphabet['N'])
        state3 = MooreState("Q3", output_alphabet['N'])
        state4 = MooreState("Q4", output_alphabet['Y'])

        state0.add_transition(input_alphabet['0'], state1)
        state0.add_transition(input_alphabet['1'], state0)
        state1.add_transition(input_alphabet['0'], state1)
        state1.add_transition(input_alphabet['1'], state2)
        state2.add_transition(input_alphabet['0'], state3)
        state2.add_transition(input_alphabet['1'], state0)
        state3.add_transition(input_alphabet['0'], state1)
        state3.add_transition(input_alphabet['1'], state4)
        state4.add_transition(input_alphabet['0'], state3)
        state4.add_transition(input_alphabet['1'], state0)

        hole_state = MooreState("Hole")

        return MooreMachineAutomaton(input_alphabet, output_alphabet, state0,
                                        set([state0, state1, state2, state3, state4]),
                                        MooreMachineComparison(),
                                        name="Sample Moore Machine 2",
                                        hole=hole_state)



import sys
import uuid
from typing import Any
from pythautomata.abstract.finite_automaton import FiniteAutomaton

from pythautomata.abstract.model import Model
from pythautomata.base_types.moore_state import MooreState
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.symbol import Symbol

from pythautomata.exceptions.unknown_symbols_exception import UnknownSymbolsException
from pythautomata.exceptions.non_deterministic_states_exception import NonDeterministicStatesException
from pythautomata.model_exporters.standard_exporters.moore_standard_dot_exporting_strategy import MooreStandardDotExportingStrategy
from pythautomata.model_exporters.dot_exporters.moore_dot_exporting_strategy import MooreDotExportingStrategy

class MooreMachineAutomaton(Model, FiniteAutomaton):
    """
    Implementation of Moore Machines.

    Attributes
    ----------
    states: set[MooreState]
        Set containing the Moore Machines states
    initial_state: State
        Initial state of the Moore Machine. Also included in "states"
    input_alphabet: Alphabet
        Set of Symbols that defines the Symbols that can be used as inputs
    output_alphabet: Alphabet
        Set of Symbols that defines the Symbols that can be used as output 

    Definition of Moore Machines is based on:
        Author, Georgios Giantamidis, Author, Stavros Tripakis (2016). 
            Learning Moore Machines from Input-Output Traces
    Link del artículo: https://arxiv.org/pdf/1605.07805.pdf
    """

    def __init__(self, input_alphabet: Alphabet, output_alphabet: Alphabet,
                 initial_state: MooreState, states: set[MooreState],
                 comparator, name: str = None,
                 exportingStrategies: list = [
                     MooreDotExportingStrategy(), MooreStandardDotExportingStrategy()],
                 hole: MooreState = MooreState('\u22A5'),
                 calculate_access_strings: bool = False):

        self.states = states
        for state in self.states:
            self._verify_state(state, input_alphabet, output_alphabet)
            state.add_hole_transition(hole)

        self._name = 'Moore Machine - ' + \
            str(uuid.uuid4().hex) if name is None else name
        self._alphabet = input_alphabet
        self._output_alphabet = output_alphabet
        self.initial_state = initial_state
        self._set_hole(hole)
        self._exporting_strategies = exportingStrategies
        self._comparator = comparator
        self._actual_state = initial_state
        super(MooreMachineAutomaton, self).__init__(comparator, calculate_access_strings)

    @property
    def name(self):
        return self._name

    @property
    def alphabet(self) -> Alphabet:
        return self._alphabet
    
    @property
    def output_alphabet(self) -> Alphabet:
        return self._output_alphabet

    def step(self, symbol):
        self._actual_state = self._actual_state.next_state_for(symbol)

        return self._actual_state.value

    def reset(self):
        self._actual_state = self.initial_state

    def last_symbol(self, sequence: Sequence) -> Symbol:
        actual_state = self.initial_state
        for symbol in sequence.value:
            actual_state = actual_state.next_state_for(symbol)
        return actual_state.value

    def process_query(self, sequence: Sequence) -> Symbol:
        return self.last_symbol(sequence)

    def transduce(self, sequence: Sequence) -> Sequence:
        actual_state = self.initial_state
        output = Sequence()
        output.append(actual_state.value)
        for symbol in sequence.value:
            actual_state = actual_state.next_state_for(symbol)
            output.append(actual_state.value)
        return output

    @property
    def initial_states(self) -> frozenset:
        return frozenset([self.initial_state])

    @property
    def hole(self):
        return self._hole

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, MooreMachineAutomaton) and self._comparator.are_equivalent(self, other)

    def export(self, path=None) -> None:
        for strategy in self._exporting_strategies:
            try:
                strategy.export(self, path)
            except:
                print("Unexpected exception when exporting " +
                      str(self._name) + ": " + str(sys.exc_info()[0]))

    def _set_hole(self, hole: MooreState) -> None:
        self._hole = hole
        # hole's hole state is itself
        self._hole.add_hole_transition(self.hole)

    def _verify_state(self, state: MooreState, alphabet: Alphabet, outputAlphabet: Alphabet) -> None:
        self._verify_symbols_in_output_alphabet(state, outputAlphabet)
        self._verify_transition_symbols_in_alphabet(
            state.transitions, alphabet)
        self._verify_state_is_deterministic(state)

    def _verify_transition_symbols_in_alphabet(self, transitions: dict[Symbol, set['MooreState']], alphabet: Alphabet) -> \
            None:
        if not all(symbol in alphabet for symbol in transitions):
            raise UnknownSymbolsException()

    def _verify_symbols_in_output_alphabet(self, state: MooreState, outputAlphabet: Alphabet) -> \
            None:
        if not state.value in outputAlphabet:
            raise UnknownSymbolsException()

    def _verify_state_is_deterministic(self, state: MooreState) -> None:
        if not state.is_deterministic:
            raise NonDeterministicStatesException()



from unittest import TestCase

from pythautomata.automata_definitions.sample_mealy_machines import SampleMealyMachines
from pythautomata.model_exporters.standard_exporters.mealy_standard_dot_exporting_strategy import MealyStandardDotExportingStrategy


class TestMealyExporting(TestCase):

    def test_1(self):
        mealy = SampleMealyMachines.get_3_states_mealy_machine()
        MealyStandardDotExportingStrategy().export(mealy, "./output_models/tests")

    def test_2(self):
        mealy = SampleMealyMachines.get_tomitas_automaton_1()
        MealyStandardDotExportingStrategy().export(mealy, "./output_models/tests")



from genericpath import isdir
from os import makedirs
from pathlib import Path
from graphviz import Digraph

from pythautomata.abstract.model_exporting_strategy import ModelExportingStrategy


class MealyStandardDotExportingStrategy(ModelExportingStrategy):
    def export(self, model, path=None, encoding=None):
        graph = self.create_graph(model)
        path = self.get_path_for(path, model)
        path = str(path) + '.dot'

        dot_code = graph.source

        with open(path, "w+", encoding=encoding) as f:
            f.write(dot_code)

    def create_graph(self, model):
        graph = Digraph('mealy-machine')

        graph.node('__start0', shape='none', label='')
        initialState = model.initial_state
        graph.edge('__start0', initialState.name)

        for state in model.states:
            graph.node(state.name, shape='circle', label=state.name)

        for state in model.states:
            for symbol, state_set in state.transitions.items():
                for destinationState in state_set:
                    label = "%s/%s" % (symbol, state.outputs[symbol])
                    graph.edge(state.name, destinationState.name, label)

        return graph



from unittest import TestCase

from pythautomata.automata_definitions.tomitas_grammars import TomitasGrammars
from pythautomata.other_models.composed_boolean_model import ComposedBooleanModel


class TestComposedBooleanModel(TestCase):

    def test_instantiate_composed_boolean_model_1(self):
        automata1 = TomitasGrammars.get_automaton_1()
        automata2 = TomitasGrammars.get_automaton_2()
        ComposedBooleanModel(
            [automata1, automata2], automata1.alphabet, lambda x, y: (x and not y))
        self.assertTrue(True)

    def test_instantiate_composed_boolean_model_exception(self):
        automata1 = TomitasGrammars.get_automaton_1()
        automata2 = TomitasGrammars.get_automaton_2()
        with self.assertRaises(TypeError):
            ComposedBooleanModel(
                [automata1, automata2], automata1.alphabet, lambda x, y, z: (x and not y or z))



from abc import ABC, abstractmethod, abstractproperty
from typing import Any


class Symbol(ABC):
    """Abstract class representing symbols.
    """
    @abstractproperty
    def value(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def copy(self) -> 'Symbol':
        raise NotImplementedError

    @abstractmethod
    def add_to_value(self, to_add: int) -> None:
        raise NotImplementedError

    def __hash__(self):
        return hash(self.value)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return self.value

    def __eq__(self, other):
        return isinstance(other, Symbol) and self.value == other.value

    def __lt__(self, other):
        return self.value < other.value

    def __gt__(self, other):
        return self.value > other.value

    def __le__(self, other):
        return self < other or self == other

    def __ge__(self, other):
        return self > other or self == other


class SymbolStr(Symbol):
    """Symbol specification using string as a representation.
    """

    def __init__(self, value: str):
        self._value = value

    def copy(self) -> 'SymbolStr':
        return SymbolStr(self._value)

    def add_to_value(self, to_add: int) -> None:
        self._value = chr(ord(self.value)+to_add)

    @property
    def value(self) -> str:
        return self._value

    def __repr__(self) -> str:
        return str(self.value)


class SymbolInt(Symbol):
    """Symbol specification using integer as a representation.
    """

    def __init__(self, value: int):
        self._value = value

    def copy(self) -> 'SymbolInt':
        return SymbolInt(self._value)

    def add_to_value(self, to_add: int) -> None:
        self._value += to_add

    @property
    def value(self) -> int:
        return self._value

    def __repr__(self) -> str:
        return str(self.value)


class SymbolInfinity(Symbol):
    """Symbol that represents infinity, not part of any alphabet. Used on interval guards
    """
    @property
    def value(self):
        return self

    def copy(self):
        return SymbolInfinity()

    def add_to_value(self, to_add: int) -> None:
        return

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return u"\u221E"

    def __eq__(self, other):
        return False

    def __lt__(self, other):
        return False

    def __gt__(self, other):
        return True


class SymbolNegativeInfinity(Symbol):
    """Symbol that represents negative infinity, not part of any alphabet. Used on interval guards
    """
    @property
    def value(self):
        return self

    def copy(self):
        return SymbolNegativeInfinity()

    def add_to_value(self, to_add: int) -> None:
        return

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return u"-\u221E"

    def __eq__(self, other):
        return False

    def __lt__(self, other):
        return True

    def __gt__(self, other):
        return False



from pythautomata.model_comparators.wfa_comparison_strategy import WFAComparator
from pythautomata.utilities import pdfa_utils
from pythautomata.utilities.probability_partitioner import ProbabilityPartitioner
from pythautomata.automata.wheighted_automaton_definition.weighted_state import WeightedState


class WFAPartitionComparator(WFAComparator):
    """
    Class containing a WFA Comparator Strategy based on partitioning of next symbol distributions.

    Methods
    -------   
    are_equivalent: bool
        returns true iif wfa1 is equivalent to wfa2 according to arg max criterion

    get_counterexample_between: Sequence
        returns a Sequence where the arg max of next token weights differ
    """

    def __init__(self, partitioner: ProbabilityPartitioner) -> None:
        super().__init__()
        self._partitioner = partitioner

    def _get_probability_distribution(self, state: WeightedState):
        _, probs, _ = state.get_all_symbol_weights()
        return probs

    def _states_are_equivalent(self, state1, state2, alphabet, ignore_initial_weight=True):
        r1 = state1.initial_weight == state2.initial_weight
        r2 = self._partitioner.are_in_same_partition(self._get_probability_distribution(
            state1), self._get_probability_distribution(state2))
        return (r1 or ignore_initial_weight) and r2

    def next_tokens_equivalent_output(self, observation1, observation2) -> bool:
        return self._partitioner.are_in_same_partition(observation1, observation2)

    def equivalent_output(self, observation1, observation2) -> bool:
        raise NotImplemented
        return pdfa_utils.have_same_argmax(observation1, observation2)

    def equivalent_values(self, value1, value2):
        raise NotImplemented
        return abs(value1 - value2) <= self.tolerance



import random
from pythautomata.automata.moore_machine_automaton import MooreMachineAutomaton
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.moore_state import MooreState
from pythautomata.model_comparators.moore_machine_comparison_strategy import MooreMachineComparisonStrategy
from pythautomata.model_exporters.image_exporters.image_exporting_strategy import ImageExportingStrategy
from pythautomata.model_exporters.dot_exporters.moore_dot_exporting_strategy import MooreDotExportingStrategy


def generate_moore_machine(input_alphabet: Alphabet, output_alphabet: Alphabet, number_of_states: int = 200, seed: int = None, exporting_strategies: list = [ImageExportingStrategy(MooreDotExportingStrategy(), "pdf")]) -> MooreMachineAutomaton:
    """
    Function returning a randomly generated Moore Machine.

    Args:
        input_alphabet (Alphabet): Moore Machine input alphabet.
        output_alphabet (Alphabet): Moore Machine output alphabet.
        number_of_states (int): Number of states of the generated Moore Machine. Defaults to 200.
        seed (int): Seed for the random number generator. Defaults to None.
        exporting_strategies (list, optional): List of strategies to export the generated Moore Machine. Defaults to [ImageExportingMMStrategy()].

    Returns:
        MooreMachineAutomaton: Random Moore Machine.
    """
    if seed is not None:
        random.seed(seed)
    states = _generate_states(number_of_states, output_alphabet)
    _add_moore_machine_transitions_to_states(states, input_alphabet.symbols)
    initial_state = next(iter(states))
    states = _remove_unreachable_states(initial_state, input_alphabet.symbols)
    comparator = MooreMachineComparisonStrategy()
    return MooreMachineAutomaton(input_alphabet, output_alphabet, initial_state, states, comparator=comparator, exportingStrategies=exporting_strategies)


def _generate_states(number_of_states, output_alphabet):
    states = []
    for index in range(number_of_states):
        generated_state = MooreState(
            # do a sample of size 1 and get the first element
            str(index), random.sample(output_alphabet.symbols, 1).pop())
        states.append(generated_state)
    return states


def _add_moore_machine_transitions_to_states(states, symbols):
    for state in states:
        for symbol in symbols:
            random_state = random.choice(states)
            state.add_transition(symbol, random_state)


def _remove_unreachable_states(initial_state, symbols):
    reachable_states = _get_reachable_states_from(initial_state, symbols)
    return reachable_states


def _get_reachable_states_from(initial_state, symbols):
    states_to_visit = [initial_state]
    visited_states = []
    while len(states_to_visit) > 0:
        state = states_to_visit.pop()
        visited_states.append(state)
        for symbol in symbols:
            next_states = state.next_states_for(symbol)
            for next_state in next_states:
                if next_state not in visited_states and next_state not in states_to_visit:
                    states_to_visit.append(next_state)
    return set(visited_states)



from pythautomata.automata.deterministic_finite_automaton import \
    DeterministicFiniteAutomaton
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.state import State
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.model_comparators.dfa_comparison_strategy import \
    DFAComparisonStrategy as DFAComparator

binaryAlphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
zero = binaryAlphabet['0']
one = binaryAlphabet['1']


class TomitasGrammars:
    """
    Class containing automata from paper:    
    Tomita, M. (1982). 
    Dynamic Construction of Finite Automata from examples using Hill-climbing. 
    Proceedings of the Fourth Annual Conference of the Cognitive Science Society 
    (p./pp. 105--108), Ann Arbor, Michigan.

    Methods
    -------
    get_all_automata: list(DeterministicFiniteAutomaton)
        returns a list containing all automata defined in this class

    get_automaton_1: DeterministicFiniteAutomaton
        returns the automaton 1 from the paper 

    get_automaton_2: DeterministicFiniteAutomaton
        returns the automaton 2 from the paper

    get_automaton_3: DeterministicFiniteAutomaton
        returns the automaton 3 from the paper

    get_automaton_4: DeterministicFiniteAutomaton
        returns the automaton 4 from the paper

    get_automaton_5: DeterministicFiniteAutomaton
        returns the automaton 5 from the paper

    get_automaton_6: DeterministicFiniteAutomaton
        returns the automaton 6 from the paper

    get_automaton_7: DeterministicFiniteAutomaton
        returns the automaton 7 from the paper
    """

    @staticmethod
    def get_all_automata(calculate_access_strings: bool = False):
        """
        Method returning a list of all automata of the class

        Returns
        -------
        list(DeterministicFiniteAutomaton)
            all automata defined in the class
        """
        return [
            TomitasGrammars.get_automaton_1(calculate_access_strings),
            TomitasGrammars.get_automaton_2(calculate_access_strings),
            TomitasGrammars.get_automaton_3(calculate_access_strings),
            TomitasGrammars.get_automaton_4(calculate_access_strings),
            TomitasGrammars.get_automaton_5(calculate_access_strings),
            TomitasGrammars.get_automaton_6(calculate_access_strings),
            TomitasGrammars.get_automaton_7(calculate_access_strings)
        ]

    @staticmethod
    def get_automaton_1(calculate_access_strings: bool = False):
        """
        Method with specification of the automaton 1 from the paper

        The first Tomita's grammar is defined as the regular expresion: 1*

        Returns
        -------
        DeterministicFiniteAutomaton
            automaton 1 from the paper
        """
        stateA = State("State A", True)
        stateB = State("State B")
        stateA.add_transition(one, stateA)
        stateA.add_transition(zero, stateB)
        stateB.add_transition(one, stateB)
        stateB.add_transition(zero, stateB)

        comparator = DFAComparator()
        return DeterministicFiniteAutomaton(binaryAlphabet, stateA,
                                            set([stateA, stateB]), comparator, 
                                            "Tomita's grammar 1 automaton",
                                            calculate_access_strings=calculate_access_strings)

    @staticmethod
    def get_automaton_2(calculate_access_strings: bool = False):
        """
        Method with specification of the automaton 2 from the paper

        The second Tomita's grammar is defined as the regular expresion: (10)*

        Returns
        -------
        DeterministicFiniteAutomaton
            automaton 2 from the paper
        """

        stateA = State("State A", True)
        stateB = State("State B")
        stateC = State("State C")
        stateA.add_transition(one, stateB)
        stateA.add_transition(zero, stateC)
        stateB.add_transition(one, stateC)
        stateB.add_transition(zero, stateA)
        stateC.add_transition(one, stateC)
        stateC.add_transition(zero, stateC)

        comparator = DFAComparator()
        return DeterministicFiniteAutomaton(binaryAlphabet, stateA,
                                            set([stateA, stateB, stateC]), comparator, 
                                            "Tomita's grammar 2 automaton",
                                            calculate_access_strings=calculate_access_strings)

    @staticmethod
    def get_automaton_3(calculate_access_strings: bool = False):
        """
        Method with specification of the automaton 3 from the paper

        The third Tomita's grammar recognizes strings that don't contain the regular expresion (1^(2n+1), 0^(2m+1)) as substring.

        Returns
        -------
        DeterministicFiniteAutomaton
            automaton 3 from the paper
        """
        stateQ1 = State("State 0", True)
        stateQ2 = State("State 1", True)
        stateQ3 = State("State 2", True)
        stateQ4 = State("State 3")
        stateQ5 = State("State 4")
        stateQ1.add_transition(one, stateQ2)
        stateQ1.add_transition(zero, stateQ1)
        stateQ2.add_transition(one, stateQ1)
        stateQ2.add_transition(zero, stateQ4)
        stateQ3.add_transition(one, stateQ2)
        stateQ3.add_transition(zero, stateQ4)
        stateQ4.add_transition(one, stateQ5)
        stateQ4.add_transition(zero, stateQ3)
        stateQ5.add_transition(one, stateQ5)
        stateQ5.add_transition(zero, stateQ5)

        comparator = DFAComparator()
        return DeterministicFiniteAutomaton(binaryAlphabet, stateQ1,
                                            set([stateQ1, stateQ2,
                                                stateQ3, stateQ4, stateQ5]),
                                            comparator,
                                            "Tomita's grammar 3 automaton",
                                            calculate_access_strings=calculate_access_strings)

    @staticmethod
    def get_automaton_4(calculate_access_strings: bool = False):
        """
        Method with specification of the automaton 4 from the paper

        The fourth Tomita grammar recognizes strings that don't contain the regular expresion 000 as substring.

        Returns
        -------
        DeterministicFiniteAutomaton
            automaton 4 from the paper
        """

        stateQ1 = State("State 0", True)
        stateQ2 = State("State 1", True)
        stateQ3 = State("State 2", True)
        stateQ4 = State("State 3")
        stateQ1.add_transition(one, stateQ1)
        stateQ1.add_transition(zero, stateQ2)
        stateQ2.add_transition(one, stateQ1)
        stateQ2.add_transition(zero, stateQ3)
        stateQ3.add_transition(one, stateQ1)
        stateQ3.add_transition(zero, stateQ4)
        stateQ4.add_transition(one, stateQ4)
        stateQ4.add_transition(zero, stateQ4)

        comparator = DFAComparator()
        return DeterministicFiniteAutomaton(binaryAlphabet, stateQ1,
                                            set([stateQ1, stateQ2, stateQ3, stateQ4]), comparator,
                                            "Tomita's grammar 4 automaton",
                                            calculate_access_strings=calculate_access_strings)

    @staticmethod
    def get_automaton_5(calculate_access_strings: bool = False):
        """
        Method with specification of the automaton 5 from the paper

        The fifth Tomita's grammar recognizes strings that have an even ammount of 01 and 10

        Returns
        -------
        DeterministicFiniteAutomaton
            automaton 5 from the paper
        """
        stateQ1 = State("State 0", True)
        stateQ2 = State("State 1")
        stateQ3 = State("State 2")
        stateQ4 = State("State 3")
        stateQ1.add_transition(one, stateQ2)
        stateQ1.add_transition(zero, stateQ3)
        stateQ2.add_transition(one, stateQ1)
        stateQ2.add_transition(zero, stateQ4)
        stateQ3.add_transition(one, stateQ4)
        stateQ3.add_transition(zero, stateQ1)
        stateQ4.add_transition(one, stateQ3)
        stateQ4.add_transition(zero, stateQ2)

        comparator = DFAComparator()
        return DeterministicFiniteAutomaton(binaryAlphabet, stateQ1,
                                            set([stateQ1, stateQ2, stateQ3, stateQ4]), comparator,
                                            "Tomita's grammar 5 automaton",
                                            calculate_access_strings=calculate_access_strings)

    @staticmethod
    def get_automaton_6(calculate_access_strings: bool = False):
        """
        Method with specification of the automaton 6 from the paper

        The sixth Tomita's grammar recognizes strings where (ammount of 0) - (amount of 1) = multiple of 3.

        Returns
        -------
        DeterministicFiniteAutomaton
            automaton 5 from the paper
        """
        stateQ1 = State("State 0", True)
        stateQ2 = State("State 1")
        stateQ3 = State("State 2")
        stateQ1.add_transition(one, stateQ2)
        stateQ1.add_transition(zero, stateQ3)
        stateQ2.add_transition(one, stateQ3)
        stateQ2.add_transition(zero, stateQ1)
        stateQ3.add_transition(one, stateQ1)
        stateQ3.add_transition(zero, stateQ2)

        comparator = DFAComparator()
        return DeterministicFiniteAutomaton(binaryAlphabet, stateQ1,
                                            set([stateQ1, stateQ2, stateQ3]), comparator, 
                                            "Tomita's grammar 6 automaton",
                                            calculate_access_strings=calculate_access_strings)

    @staticmethod
    def get_automaton_7(calculate_access_strings: bool = False):
        """
        Method with specification of the automaton 7 from the paper

        The seventh Tomita's grammargrammar is defined as the regular expresion: 0*1*0*1*
        Returns
        -------
        DeterministicFiniteAutomaton
            automaton 7 from the paper
        """
        stateQ1 = State("state0", True)
        stateQ2 = State("state1", True)
        stateQ3 = State("state2", True)
        stateQ4 = State("state3", True)
        stateQ5 = State("state4")
        stateQ1.add_transition(one, stateQ2)
        stateQ1.add_transition(zero, stateQ1)
        stateQ2.add_transition(one, stateQ2)
        stateQ2.add_transition(zero, stateQ3)
        stateQ3.add_transition(one, stateQ4)
        stateQ3.add_transition(zero, stateQ3)
        stateQ4.add_transition(one, stateQ4)
        stateQ4.add_transition(zero, stateQ5)
        stateQ5.add_transition(one, stateQ5)
        stateQ5.add_transition(zero, stateQ5)

        comparator = DFAComparator()
        return DeterministicFiniteAutomaton(binaryAlphabet, stateQ1,
                                            set([stateQ1, stateQ2, stateQ3,
                                                stateQ4, stateQ5]), comparator,
                                            "Tomita's grammar 7 automaton",
                                            calculate_access_strings=calculate_access_strings)



from .symbol import Symbol
from pythautomata.exceptions.none_state_exception import NoneStateException

class State:
    """Representation of NFA or DFA states.

    Attributes
    ----------
    name: str
        State name.
    is_final: bool
        Determines if the state is final.
    transitions: dict[Symbol, set['State']]
        For any given symbol represents the next state (or set of states in the case of NFA).
    hole:
        Hole state, state containing all transitions directed to itself. 
        It is used as default when a symbol is not present as transition key.
    """
    
    hole: 'State'

    def __init__(self, name: str, is_final: bool = False):
        self.name = name
        self.is_final = is_final
        self.transitions: dict[Symbol, set['State']] = {}
        self._is_deterministic: bool = True
        self.access_string = None

    @property
    def is_deterministic(self) -> bool:
        return self._is_deterministic

    def add_transition(self, symbol: Symbol, next_state: 'State') -> None:
        if next_state is None:
            raise NoneStateException()
        if symbol not in self.transitions:
            self.transitions[symbol] = set()
        else:
            self._is_deterministic = False
        self.transitions[symbol].add(next_state)

    def add_multiple_transitions(self, symbol: Symbol, next_states: list['State']) -> None:
        for next_state in next_states:
            self.add_transition(symbol, next_state)

    def next_states_for(self, symbol: Symbol) -> set['State']:
        if symbol not in self.transitions:
            return {self.hole}
        return self.transitions[symbol]

    def next_state_for(self, symbol: Symbol) -> 'State':
        next_states = list(self.next_states_for(symbol))
        assert(len(next_states) == 1)
        return next_states[0]

    def add_hole_transition(self, hole: 'State') -> None:
        self.hole = hole

    def __eq__(self, other):
        return isinstance(other, State) and self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return str(self.name) + (" (Final)" if self.is_final else " (Non-final)")



from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence
from pythautomata.utilities.sequence_generator import SequenceGenerator
from pythautomata.automata.wheighted_automaton_definition.weighted_automaton import WeightedAutomaton as WFA
import random

class GuidingWDFASequenceGenerator(SequenceGenerator):    
    def __init__(self, wfa: WFA, max_seq_length: int, random_seed: int = 21, random_stop_proba = 0.2):
        self.wfa = wfa
        self._random_stop_proba = random_stop_proba
        super().__init__(wfa.alphabet, max_seq_length, random_seed)
    
    def generate_words(self, number_of_words: int):
        result = list()
        for _ in range(number_of_words):
            word = self.generate_single_word(None)
            result.append(word)
        return result

    def _sort_valid_symbol(self, symbols, weights):
        if random.random() < self._random_stop_proba:
            return self.wfa.terminal_symbol
        selected_symbols = []
        for i,w in enumerate(weights):
            if w>0:
                selected_symbols.append(symbols[i])
        next_symbol = random.choice(selected_symbols)
        return next_symbol
    
    def generate_single_word(self, length):
        assert length is None
        word = Sequence()
        first_state = list(filter(lambda x: x.initial_weight ==
                        1, self.wfa.weighted_states))[0]
        symbols, weights, next_states = first_state.get_all_symbol_weights()
        next_symbol = self._sort_valid_symbol(symbols, weights)
        while next_symbol != self.wfa.terminal_symbol:
            word += next_symbol
            i = symbols.index(next_symbol)
            next_state = next_states[i]
            symbols, weights, next_states = next_state.get_all_symbol_weights()
            next_symbol = self._sort_valid_symbol(symbols, weights)
        return word


from unittest import TestCase

from pythautomata.automata_definitions.bollig_habermehl_kern_leucker_automata import BolligHabermehlKernLeuckerAutomata
from pythautomata.automata_definitions.omlin_giles_automata import OmlinGilesAutomata
from pythautomata.automata_definitions.other_automata import OtherAutomata
from pythautomata.automata_definitions.sample_mealy_machines import SampleMealyMachines
from pythautomata.automata_definitions.sample_nfas import SampleNFAs
from pythautomata.automata_definitions.tomitas_grammars_modifications import TomitasGrammarsMods
from pythautomata.automata_definitions.tomitas_grammars import TomitasGrammars
from pythautomata.automata_definitions.weighted_tomitas_grammars import WeightedTomitasGrammars
from pythautomata.automata_definitions.sample_moore_machines import SampleMooreMachines


class TestAutomataDefinitions(TestCase):

    def test_instantiating_all_automata(self):
        automata1 = BolligHabermehlKernLeuckerAutomata.get_all_automata()
        automata2 = OmlinGilesAutomata.get_all_automata()
        automata3 = OtherAutomata.get_all_automata()
        automata4 = SampleNFAs.get_all_automata()
        automata5 = TomitasGrammarsMods.get_all_automata()
        automata6 = TomitasGrammars.get_all_automata()
        automata7 = WeightedTomitasGrammars.get_all_automata()
        automata8 = SampleMooreMachines.get_all_automata()
        automata9 = SampleMealyMachines.get_all_mealy_machines()

        self.assertTrue(True)

    def test_bad_instantiating_automata(self):
        try:
            SampleMooreMachines.get_3_states_automaton_wrong_alphabet()
            self.assertTrue(False)
        except:
            pass

        try:
            SampleMooreMachines.get_3_states_automaton_non_deterministic()
            self.assertTrue(False)
        except:
            pass

        self.assertTrue(True)



from unittest import TestCase
from pythautomata.utilities import nicaud_mm_generator
from pythautomata.utilities import simple_mm_generator
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.symbol import SymbolStr

binaryAlphabet = Alphabet(
    frozenset([SymbolStr('0'), SymbolStr('1')]))

tripleAlphabet = Alphabet(
    frozenset([SymbolStr('a'), SymbolStr('b'), SymbolStr('c')]))


class TestMooreGenerators(TestCase):

    def test_simple_moore_generator_1(self):
        generated_moore = simple_mm_generator.generate_moore_machine(
            binaryAlphabet, tripleAlphabet, 80)
        self._assert_correctness(generated_moore)

    def test_simple_dfa_generator_1(self):
        generated_moore = simple_mm_generator.generate_moore_machine(
            binaryAlphabet, tripleAlphabet, 80)
        self._assert_correctness(generated_moore)

    def test_simple_dfa_generator_2(self):
        generated_moore = simple_mm_generator.generate_moore_machine(
            binaryAlphabet, tripleAlphabet, 49)
        self._assert_correctness(generated_moore)

    def test_simple_dfa_generator_3(self):
        generated_moore = simple_mm_generator.generate_moore_machine(
            binaryAlphabet, tripleAlphabet, 49)
        self._assert_correctness(generated_moore)

    def test_simple_dfa_generator_4(self):
        generated_moore = simple_mm_generator.generate_moore_machine(
            binaryAlphabet, tripleAlphabet, 1)
        self._assert_correctness(generated_moore)

    def test_nicaud_1(self):
        symbols = []
        for i in range(10):
            symbols.append(SymbolStr(str(i)))
        mediumAlphabet = Alphabet(frozenset(symbols))
        generated_moore = nicaud_mm_generator.generate_moore_machine(
            mediumAlphabet, tripleAlphabet, 1000)
        self._assert_correctness(generated_moore)

    def _assert_correctness(self, automaton):
        self.assertTrue(self._all_states_are_rechable(automaton))

    def _all_states_are_rechable(self, automaton):
        unrechable = automaton.states.copy()
        for state in unrechable.copy():
            for destinations in state.transitions.values():
                unrechable = unrechable - destinations
        return len(unrechable) <= 1  # Hole



from symtable import Symbol
from graphviz import Digraph

from pythautomata.abstract.model_exporting_strategy import ModelExportingStrategy


class MooreDotExportingStrategy(ModelExportingStrategy):
    def export(self, model, path=None):
        graph = self.create_graph(model)
        path = self.get_path_for(path, model)
        path = str(path) + '.dot'

        dot_code = graph.source

        with open(path, "w+", encoding="utf-8") as f:
            f.write(dot_code)

    def create_graph(self, model):
        graph = Digraph('moore_machine')
        graph.attr(rankdir='LR', size='8,5')

        graph.attr('node', shape='circle')
        for state in model.states:
            label = state.name + '/' + str(state.value)
            graph.node(state.name, label=label)
            transitions = dict()
            for symbol, destinationStates in state.transitions.items():
                for destinationState in destinationStates:
                    transitions.setdefault(
                        (state.name, destinationState.name), set())
                    transitions[(state.name, destinationState.name)].add(
                        str(symbol))

            for (state_from, state_to), symbols in transitions.items():
                label = self._get_label_for(symbols, model._alphabet)
                graph.edge(state_from, state_to, label)

        graph.attr('node', shape='point')
        nodeName = 'start'
        graph.node(nodeName)
        graph.edge(nodeName, model.initial_state.name)

        return graph

    def _get_label_for(self, symbols: list, alphabet):
        if len(symbols) == len(alphabet):
            return "Σ"
        elif len(symbols) <= len(alphabet) / 2:
            return ", ".join(sorted(symbols))
        else:
            complement = set(map(str, alphabet.symbols)) - set(symbols)
            label = ", ".join(sorted(complement))
            return f"Σ - {{{label}}}"



from .symbol import Symbol, SymbolStr


class Alphabet:
    """Set of Symbols.
    """

    @staticmethod
    def from_strings(strings: list[str]) -> 'Alphabet':
        """Create an alphabet from a list of strings.

        Parameters
        ----------
        strings : list[str]
            List of strings

        Returns
        -------
        Alphabet
            Alphabet
        """
        return Alphabet(frozenset(map(SymbolStr, strings)))

    def __init__(self, symbols: frozenset[Symbol], name: str = None):
        """Constructor

        Parameters
        ----------
        symbols : frozenset[Symbol]
            Symbols belonging to alphabet
        name : str, optional
            An optional name, by default None
        """
        self.symbols = symbols
        self.name = name

    def __getitem__(self, key: str):
        return next(symbol for symbol in self.symbols if str(symbol) == key)

    def __contains__(self, symbol: Symbol):
        return symbol in self.symbols

    def __len__(self):
        return len(self.symbols)

    def __eq__(self, other):
        return isinstance(other, Alphabet) and self.symbols == other.symbols

    def __hash__(self):
        return hash(self.symbols)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return str(self.symbols)



import unittest
from pythautomata.utilities import pdfa_metrics
from pythautomata.utilities.uniform_length_sequence_generator import UniformLengthSequenceGenerator
from pythautomata.utilities.uniform_word_sequence_generator import UniformWordSequenceGenerator
from pythautomata.automata_definitions.weighted_tomitas_grammars import WeightedTomitasGrammars


class TestPDFAMetrics(unittest.TestCase):

    def test_same_pdfa(self):
        pdfa1 = WeightedTomitasGrammars.get_automaton_3()
        pdfa2 = pdfa1
        sg = UniformLengthSequenceGenerator(pdfa1.alphabet, 20, 42)
        test_sequences = sg.generate_words(500)
        m1 = pdfa_metrics.log_probability_error(pdfa1, pdfa2, test_sequences)
        m2 = pdfa_metrics.wer_avg(pdfa1, pdfa2, test_sequences)
        m3 = pdfa_metrics.ndcg_score_avg(pdfa1, pdfa2, test_sequences)
        m4 = pdfa_metrics.out_of_partition_elements(
            pdfa1, pdfa2, test_sequences, 10)
        m5 = pdfa_metrics.out_of_tolerance_elements(
            pdfa1, pdfa2, test_sequences, 0.1)
        m6 = pdfa_metrics.absolute_error_avg(pdfa1, pdfa2, test_sequences)
        m7 = pdfa_metrics.mean_cross_entropy(pdfa1, pdfa2, test_sequences)
        self.assertEqual(m1, 0)
        self.assertEqual(m2, 0)
        self.assertEqual(m3, 1)
        self.assertEqual(m4, 0)
        self.assertEqual(m5, 0)
        self.assertEqual(m6, 0)
        self.assertNotEqual(m7, 0)

    def test_different_pdfa(self):
        pdfa1 = WeightedTomitasGrammars.get_automaton_3()
        pdfa2 = WeightedTomitasGrammars.get_automaton_7()
        sg = UniformLengthSequenceGenerator(pdfa1.alphabet, 20, 42)
        test_sequences = sg.generate_words(500)
        m1 = pdfa_metrics.log_probability_error(pdfa1, pdfa2, test_sequences)
        m2 = pdfa_metrics.wer_avg(pdfa1, pdfa2, test_sequences)
        m3 = pdfa_metrics.ndcg_score_avg(pdfa1, pdfa2, test_sequences)
        m4 = pdfa_metrics.out_of_partition_elements(
            pdfa1, pdfa2, test_sequences, 10)
        m5 = pdfa_metrics.out_of_tolerance_elements(
            pdfa1, pdfa2, test_sequences, 0.1)

        m6 = pdfa_metrics.absolute_error_avg(pdfa1, pdfa2, test_sequences)
        m7 = pdfa_metrics.mean_cross_entropy(pdfa1, pdfa2, test_sequences)
        self.assertNotEqual(m1, 0)
        self.assertNotEqual(m2, 0)
        self.assertNotEqual(m3, 1)
        self.assertNotEqual(m4, 0)
        self.assertNotEqual(m5, 0)
        self.assertNotEqual(m6, 0)
        self.assertNotEqual(m6, 0)
        
    def test_same_pdfa_uniform_word(self):
        pdfa1 = WeightedTomitasGrammars.get_automaton_3()
        pdfa2 = pdfa1
        sg = UniformWordSequenceGenerator(pdfa1.alphabet, min_seq_length=5, max_seq_length=20, random_seed=42)
        test_sequences = sg.generate_words(500)
        m1 = pdfa_metrics.log_probability_error(pdfa1, pdfa2, test_sequences)
        m2 = pdfa_metrics.wer_avg(pdfa1, pdfa2, test_sequences)
        m3 = pdfa_metrics.ndcg_score_avg(pdfa1, pdfa2, test_sequences)
        m4 = pdfa_metrics.out_of_partition_elements(
            pdfa1, pdfa2, test_sequences, 10)
        m5 = pdfa_metrics.out_of_tolerance_elements(
            pdfa1, pdfa2, test_sequences, 0.1)
        m6 = pdfa_metrics.absolute_error_avg(pdfa1, pdfa2, test_sequences)
        m7 = pdfa_metrics.mean_cross_entropy(pdfa1, pdfa2, test_sequences)
        self.assertEqual(m1, 0)
        self.assertEqual(m2, 0)
        self.assertEqual(m3, 1)
        self.assertEqual(m4, 0)
        self.assertEqual(m5, 0)
        self.assertEqual(m6, 0)
        self.assertNotEqual(m7, 0)

    def test_different_pdfa_uniform_word(self):
        pdfa1 = WeightedTomitasGrammars.get_automaton_3()
        pdfa2 = WeightedTomitasGrammars.get_automaton_7()
        sg = UniformWordSequenceGenerator(pdfa1.alphabet, min_seq_length=5, max_seq_length=20, random_seed=42)
        test_sequences = sg.generate_words(500)
        m1 = pdfa_metrics.log_probability_error(pdfa1, pdfa2, test_sequences)
        m2 = pdfa_metrics.wer_avg(pdfa1, pdfa2, test_sequences)
        m3 = pdfa_metrics.ndcg_score_avg(pdfa1, pdfa2, test_sequences)
        m4 = pdfa_metrics.out_of_partition_elements(
            pdfa1, pdfa2, test_sequences, 10)
        m5 = pdfa_metrics.out_of_tolerance_elements(
            pdfa1, pdfa2, test_sequences, 0.1)

        m6 = pdfa_metrics.absolute_error_avg(pdfa1, pdfa2, test_sequences)
        m7 = pdfa_metrics.mean_cross_entropy(pdfa1, pdfa2, test_sequences)
        self.assertNotEqual(m1, 0)
        self.assertNotEqual(m2, 0)
        self.assertNotEqual(m3, 1)
        self.assertNotEqual(m4, 0)
        self.assertNotEqual(m5, 0)
        self.assertNotEqual(m6, 0)
        self.assertNotEqual(m6, 0)
